# -*- coding: utf-8 -*-
"""
Sample parsing functions, adapt these to make your own fyrd queue parser.
"""
import os as _os
import re as _re
import sys as _sys
import pwd as _pwd     # Used to get usernames for queue
from subprocess import CalledProcessError as _CalledProcessError

from .. import run as _run
from .. import logme as _logme
from .. import ClusterError as _ClusterError
from .. import script_runners as _scrpts
from .. import submission_scripts as _sscrpt
_Script = _sscrpt.Script


# This will be prepended to parsed text in script files
# e.g. #SBATCH for slurm
PREFIX = ''
# This will be appended to job submission scripts, e.g. '.qsub' for torque or
# '.sbatch' for slurm
SUFFIX = '.sh'


###############################################################################
#                             Functionality Test                              #
###############################################################################


def queue_test(warn=True):
    """Check that this batch system can be used.

    Parameters
    ----------
    warn : bool
        log a warning on fail

    Returns
    -------
    batch_system_functional : bool
    """
    log_level = 'error' if warn else 'debug'
    pass


###############################################################################
#                           Normalization Functions                           #
###############################################################################


def normalize_job_id(job_id):
    """Convert the job id into job_id, array_id."""
    pass


def normalize_state(state):
    """Convert state into standadized (slurm style) state."""
    return state


###############################################################################
#                               Job Submission                                #
###############################################################################


def gen_scripts(job_object, command, args, precmd, modstr):
    """Build the submission script objects.

    This script should almost certainly work by formatting `_scrpts.CMND_RUNNER_TRACK`.
    The result should be a script that can be executed on a node by the batch system.
    The format of the output is important, which is why `_scrpts.CMND_RUNNER_TRACK`
    should be used; if it is not used, then be sure to copy the format of the outfile
    in that script.

    Parameters
    ---------
    job_object : fyrd.job.Job
    command : str
        Command to execute
    args : list
        List of additional arguments
    precmd : str
        String from options_to_string() to add at the top of the file, should
        contain batch system directives
    modstr : str
        String to add after precmd, should contain module directives.

    Returns
    -------
    fyrd.script_runners.Script
        The submission script
    fyrd.script_runners.Script, or None if unneeded
        As execution script that will be called by the submission script,
        optional
    """
    scrpt = _os.path.join(
        job_object.scriptpath,
        '{0}.{1}.{2}'.format(job_object.name, job_object.suffix, SUFFIX)
    )

    sub_script = _scrpts.CMND_RUNNER_TRACK.format(
        precmd=precmd, usedir=job_object.runpath, name=job_object.name,
        command=command
    )
    return _Script(script=sub_script, file_name=scrpt), None


def submit(file_name, dependencies=None, job=None, args=None, kwds=None):
    """Submit any file with dependencies.

    If your batch system does not handle dependencies, then raise a
    NotImplemented error if dependencies are passed.

    Parameters
    ----------
    file_name : str
        Path to an existing file
    dependencies : list, optional
        List of dependencies
    job : fyrd.job.Job, optional, not required
        A job object for the calling job
    args : list, optional, not required
        A list of additional command line arguments to pass when submitting
    kwds : dict or str, optional, not required
        A dictionary of keyword arguments to parse with options_to_string, or
        a string of option:value,option,option:value,....

    Returns
    -------
    job_id : str
    """
    pass


###############################################################################
#                               Job Management                                #
###############################################################################


def kill(job_ids):
    """Terminate all jobs in job_ids.

    Parameters
    ----------
    job_ids : list or str
        A list of valid job ids or a single valid job id

    Returns
    -------
    success : bool
    """
    pass


###############################################################################
#                                Queue Parsing                                #
###############################################################################


def queue_parser(user=None, partition=None):
    """Iterator for queue parsing.

    Parameters
    ----------
    user : str, optional
        User name to pass to qstat to filter queue with
    partiton : str, optional
        Partition to filter the queue with

    Yields
    ------
    job_id : str
    array_id : str or None
    name : str
    userid : str
    partition : str
    state :str
    nodelist : list
    numnodes : int
    cntpernode : int or None
    exit_code : int or Nonw
    """
    pass


def parse_strange_options(option_dict):
    """Parse all options that cannot be handled by the regular function.

    Parameters
    ----------
    option_dict : dict
        All keyword arguments passed by the user that are not already defined
        in the Job object

    Returns
    -------
    list
        A list of strings to be added at the top of the script file
    dict
        Altered version of option_dict with all options that can't be handled
        by `fyrd.batch_systems.options.option_to_string()` removed.
    list
        A list of command line arguments to pass straight to the submit
        function
    """
    pass



# -*- coding: utf-8 -*-
"""
Logging with timestamps and optional log files.

Print a timestamped message to a logfile, STDERR, or STDOUT.

If STDERR or STDOUT are used, colored flags are added.  Colored flags are INFO,
WARNINING, ERROR, or CRITICAL.

It is possible to write to both logfile and STDOUT/STDERR using the also_write
argument.

If level is 'error' or 'critical', error is written to STDERR unless also_write
== -1

MIN_LEVEL can also be provided, logs will only print if vlevel > MIN_LEVEL.
Level order: critical>error>warn>info>debug>verbose

Usage::

    import logme as lm
    lm.log("Screw up!", <outfile>,
        level='debug'|'info'|'warn'|'error'|'normal',
        also_write='stderr'|'stdout')

Example::

    lm.log('Hi')
    Prints: 20160223 11:46:24.969 | INFO --> Hi
    lm.log('Hi', level='debug')
    Prints nothing
    lm.MIN_LEVEL = 'debug'
    lm.log('Hi', level='debug')
    Prints: 20160223 11:46:24.969 | DEBUG --> Hi

Note: Uses terminal colors and STDERR, not compatible with non-unix systems
"""
import sys
import gzip
import bz2
import logging
from datetime import datetime as dt

__all__ = ['log', 'MIN_LEVEL', 'LOGFILE']

###################################
#  Constants for printing colors  #
###################################

WHITE  = '\033[97m'
YELLOW = '\033[93m'
RED    = '\033[91m'
BOLD   = '\033[1m'
ENDC   = '\033[0m'

MIN_LEVEL = 'info'
LOGFILE   = sys.stderr


def log(message, level='info', logfile=None, also_write=None,
        min_level=None, kind=None):
    """Print a string to logfile.

    Levels display as::

        verbose:  <timestamp> VERBOSE -->
        debug:    <timestamp> DEBUG -->
        info:     <timestamp> INFO -->
        warn:     <timestamp> WARNING -->
        error:    <timestamp> ERROR -->
        critical: <timestamp> CRITICAL -->

    Parameters
    ----------
    message : str, optional
        The message to print.
    logfile : file or logging object, optional
        Optional file to log to, defaults to STDERR. Can provide a logging
        object
    level : {'debug', 'info', 'warn', 'error', 'normal'}, optional
        Will only print if level > MIN_LEVEL
    also_write : {'stdout', 'stderr'}, optional
        Print to STDOUT or STDERR also.
        These only have an effect if the output is not already set to the same
        device.
    min_level : str, deprecated
        Retained for backwards compatibility, min_level should be set using the
        logme.MIN_LEVEL constant.
    kind : str, deprecated
        synonym for level, kept to retain backwards compatibility
    """
    stdout = False
    stderr = False
    message = str(message)

    if not logfile:
        logfile = LOGFILE

    if kind:
        level = kind

    min_level = min_level if min_level else MIN_LEVEL

    # Level checking, not used with logging objects
    level_map = {'verbose': 0, 'debug': 1, 'info': 2, 'warn': 3, 'error': 4,
                 'critical': 5,
                 'v': 0, 'd': 1, 'i': 2, 'w': 3, 'e': 4, 'c': 5,
                 0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5}

    try:
        level = level_map[level]
    except KeyError:
        raise Exception('Invalid level {}'.format(level))

    try:
        min_level = level_map[min_level]
    except KeyError:
        raise Exception('Invalid min_level {}'.format(min_level))

    if level > 3:
        if also_write != -1 or also_write != 'stdout':
            also_write = 'stderr'

    # Attempt to handle all file types
    if isinstance(logfile, (logging.RootLogger, logging.Logger)):
        _logit(message, logfile, level, color=False, min_level=min_level)
    elif isinstance(logfile, str):
        with _open_zipped(logfile, 'a') as outfile:
            _logit(message, outfile, level, color=False, min_level=min_level)
    elif str(getattr(logfile, 'name')).strip('<>') == 'stdout':
        _logit(message, logfile, level, color=True, min_level=min_level)
        stdout = True
    elif str(getattr(logfile, 'name')).strip('<>') == 'stderr':
        _logit(message, logfile, level, color=True, min_level=min_level)
        stderr = True
    elif getattr(logfile, 'closed'):
        with _open_zipped(logfile.name, 'a') as outfile:
            _logit(message, outfile, level, color=False, min_level=min_level)
    else:
        _logit(message, logfile, level, color=False, min_level=min_level)

    # Also print to stdout or stderr if requested
    if also_write == 'stdout' and not stdout:
        _logit(message, sys.stdout, level, color=True, min_level=min_level)
    elif also_write == 'stderr' and not stderr:
        _logit(message, sys.stdout, level, color=True, min_level=min_level)


def clear(infile):
    """Truncate a file."""
    open(infile, 'w').close()


###############################################################################
#                             A Logging Exception                             #
###############################################################################


class LoggingException(Exception):

    """Log a critical message with logme and also raise."""

    def __init__(self, message, logfile=None):
        """Log message as critical, raise with first line."""
        args = {'kind': 'critical'}
        if logfile:
            args.update({'logfile': logfile})
        # Log with logme
        log(message, **args)
        # Raise with the first line of the log
        message = message.split('\n')[0]
        super(LoggingException, self).__init__(message)


###############################################################################
#                              Private Functions                              #
###############################################################################


def _logit(message, output, level, color=False, min_level=None):
    """Write message to file either with color or not.

    output must be filehandle or logging object.
    """
    now = dt.now()
    timestamp = "{0}.{1:<3}".format(now.strftime("%Y%m%d %H:%M:%S"),
                                    str(int(now.microsecond/1000)))

    flag_map  = {0: 'VERBOSE', 1: 'DEBUG', 2: 'INFO', 3: 'WARNING', 4: 'ERROR',
                 5: 'CRITICAL'}

    flag = flag_map[level]
    flag_len = len('{0} | {1} --> '.format(timestamp, flag)) - 2

    if color:
        flag = _color(flag)

    if isinstance(output, (logging.RootLogger, logging.Logger)):
        message = ' {} --> {}'.format(timestamp, message)
        if level == 0:
            output.debug(message)
        if level == 1:
            output.debug(message)
        if level == 2:
            output.info(message)
        if level == 3:
            output.warning(message)
        if level == 4:
            output.error(message)
        if level == 5:
            output.critical(message)
    else:
        # Check min_level before proceeding
        if level < min_level:
            return

        # Format multiline message
        lines = message.split('\n')
        if len(lines) != 1:
            message = lines[0] + '\n'
            lines = lines[1:]
            for line in lines:
                message = message + (''.ljust(flag_len, '-') + '> ' +
                                     line + '\n')
        output.write('{0} | {1} --> {2}\n'.format(timestamp, flag,
                                                  str(message)))


def _color(flag):
    """Return the flag with correct color codes."""
    if flag == 'VERBOSE':
        return flag
    elif flag == 'DEBUG':
        return flag
    elif flag == 'INFO':
        return BOLD + WHITE + flag + ENDC
    elif flag == 'WARNING':
        return BOLD + YELLOW + flag + ENDC
    elif flag == 'ERROR':
        return BOLD + RED + flag + ENDC
    elif flag == 'CRITICAL':
        return BOLD + RED + flag + ENDC
    else:
        raise Exception('Invalid flag type')


def _open_zipped(infile, mode='r'):
    """Return file handle of file regardless of zipped or not.

    Text mode enforced for compatibility with python2
    """
    mode   = mode[0] + 't'
    p2mode = mode
    if hasattr(infile, 'write'):
        return infile
    if isinstance(infile, str):
        if infile.endswith('.gz'):
            return gzip.open(infile, mode)
        if infile.endswith('.bz2'):
            if hasattr(bz2, 'open'):
                return bz2.open(infile, mode)
            else:
                return bz2.BZ2File(infile, p2mode)
        return open(infile, p2mode)



# -*- coding: utf-8 -*-
"""
SLURM parsing functions.
"""
import os as _os
import re as _re
import sys as _sys
import pwd as _pwd     # Used to get usernames for queue
from subprocess import CalledProcessError as _CalledProcessError

from six import text_type as _txt
from six import string_types as _str
from six import integer_types as _int

from .. import run as _run
from .. import conf as _conf
from .. import logme as _logme
from .. import ClusterError as _ClusterError
from .. import script_runners as _scrpts
from .. import submission_scripts as _sscrpt
_Script = _sscrpt.Script


PREFIX = '#SBATCH'
SUFFIX = 'sbatch'


###############################################################################
#                             Functionality Test                              #
###############################################################################


def queue_test(warn=True):
    """Check that slurm can be used.

    Just looks for sbatch and squeue.

    Parameters
    ----------
    warn : bool
        log a warning on fail

    Returns
    -------
    batch_system_functional : bool
    """
    log_level = 'error' if warn else 'debug'
    sbatch = _conf.get_option('queue', 'sbatch')
    if sbatch is not None and _os.path.dirname(sbatch) and not _run.is_exe(sbatch):
        _logme.log(
            'Cannot use slurm as sbatch path set in conf to {0}'
            .format(sbatch) + ' but that path is not an executable',
            log_level
        )
        return False
    sbatch = sbatch if sbatch else 'sbatch'
    sbatch = _run.which(sbatch) if not _os.path.dirname(sbatch) else sbatch
    if not sbatch:
        _logme.log(
            'Cannot use slurm as cannot find sbatch', log_level
        )
        return False
    qpath = _os.path.dirname(sbatch)
    squeue = _os.path.join(qpath, 'squeue')
    return _run.is_exe(squeue)


###############################################################################
#                           Normalization Functions                           #
###############################################################################


def normalize_job_id(job_id):
    """Convert the job id into job_id, array_id."""
    if '_' in job_id:
        job_id, array_id = job_id.split('_')
        job_id = job_id.strip()
        array_id = array_id.strip()
    else:
        array_id = None
    return job_id, array_id


def normalize_state(state):
    """Convert state into standadized (slurm style) state."""
    return state


###############################################################################
#                               Job Submission                                #
###############################################################################


def gen_scripts(job_object, command, args, precmd, modstr):
    """Build the submission script objects.

    Creates an exec script as well as a submission script.

    Parameters
    ---------
    job_object : fyrd.job.Job
    command : str
        Command to execute
    args : list
        List of additional arguments, not used in this script.
    precmd : str
        String from options_to_string() to add at the top of the file, should
        contain batch system directives
    modstr : str
        String to add after precmd, should contain module directives.

    Returns
    -------
    fyrd.script_runners.Script
        The submission script
    fyrd.script_runners.Script
        The execution script
    """
    scrpt = _os.path.join(
        job_object.scriptpath, '{}.{}.{}'.format(
            job_object.name, job_object.suffix, SUFFIX
        )
    )

    # We use a separate script and a single srun command to avoid
    # issues with multiple threads running at once
    exec_script  = _os.path.join(
        job_object.scriptpath, '{}.{}.script'.format(
            job_object.name, job_object.suffix
        )
    )
    exe_script   = _scrpts.CMND_RUNNER_TRACK.format(
        precmd=modstr, usedir=job_object.runpath, name=job_object.name,
        command=command
    )

    # Create the exec_script Script object
    exec_script_obj = _Script(
        script=exe_script, file_name=exec_script
    )

    ecmnd = 'srun bash {}'.format(exec_script)
    sub_script = _scrpts.SCRP_RUNNER.format(
        precmd=precmd, script=exec_script, command=ecmnd
    )

    submission_script = _Script(script=sub_script, file_name=scrpt)

    return submission_script, exec_script_obj


def submit(file_name, dependencies=None, job=None, args=None, kwds=None):
    """Submit any file with dependencies to Torque.

    Parameters
    ----------
    file_name : str
        Path to an existing file
    dependencies : list
        List of dependencies
    job : fyrd.job.Job, not implemented
        A job object for the calling job, not used by this functions
    args : list, not implemented
        A list of additional command line arguments to pass when submitting,
        not used by this function
    kwds : dict or str, not implemented
        A dictionary of keyword arguments to parse with options_to_string, or
        a string of option:value,option,option:value,.... Not used by this
        function.

    Returns
    -------
    job_id : str
    """
    _logme.log('Submitting to slurm', 'debug')
    if dependencies:
        deps = '--dependency=afterok:{}'.format(
            ':'.join([str(d) for d in dependencies]))
        args = ['sbatch', deps, file_name]
    else:
        args = ['sbatch', file_name]

    # Try to submit job 5 times
    code, stdout, stderr = _run.cmd(args, tries=5)
    if code == 0:
        job_id, _ = normalize_job_id(stdout.split(' ')[-1])
    else:
        _logme.log('sbatch failed with code {}\n'.format(code) +
                   'stdout: {}\nstderr: {}'.format(stdout, stderr),
                   'critical')
        raise _CalledProcessError(code, args, stdout, stderr)

    return job_id


###############################################################################
#                               Job Management                                #
###############################################################################


def kill(job_ids):
    """Terminate all jobs in job_ids.

    Parameters
    ----------
    job_ids : list or str
        A list of valid job ids or a single valid job id

    Returns
    -------
    success : bool
    """
    o = _run.cmd('scancel {0}'.format(' '.join(_run.listify(job_ids))),
                 tries=5)
    return o[0] == 0


###############################################################################
#                                Queue Parsing                                #
###############################################################################


def queue_parser(user=None, partition=None):
    """Iterator for slurm queues.

    Use the `squeue -O` command to get standard data across implementation,
    supplement this data with the results of `sacct`. sacct returns data only
    for the current user but retains a much longer job history. Only jobs not
    returned by squeue are added with sacct, and they are added to *the end* of
    the returned queue, i.e. *out of order with respect to the actual queue*.

    Parameters
    ----------
    user : str, optional
        User name to pass to qstat to filter queue with
    partiton : str, optional
        Partition to filter the queue with

    Yields
    ------
    job_id : str
    array_id : str or None
    name : str
    userid : str
    partition : str
    state :str
    nodelist : list
    numnodes : int
    cntpernode : int or None
    exit_code : int or Nonw
    """
    nodequery = _re.compile(r'([^\[,]+)(\[[^\[]+\])?')
    qargs = ['squeue', '-h', '-O',
             'jobid:400,arraytaskid:400,name:400,userid:400,partition:400,' +
             'state:400,nodelist:400,numnodes:400,numcpus:400,exit_code:400']
    # Parse queue info by length
    squeue = [
        tuple(
            [k[i:i+200].rstrip() for i in range(0, 4000, 400)]
        ) for k in _run.cmd(qargs)[1].split('\n')
    ]
    # SLURM sometimes clears the queue extremely fast, so we use sacct
    # to get old jobs by the current user
    qargs = ['sacct', '-p',
             '--format=jobid,jobname,user,partition,state,' +
             'nodelist,reqnodes,ncpus,exitcode']
    try:
        sacct = [tuple(i.strip(' |').split('|')) for i in
                 _run.cmd(qargs)[1].split('\n')]
        sacct = sacct[1:]
    # This command isn't super stable and we don't care that much, so I will
    # just let it die no matter what
    except Exception as e:
        if _logme.MIN_LEVEL == 'debug':
            raise e
        else:
            sacct = []

    if sacct:
        if len(sacct[0]) != 9:
            _logme.log('sacct parsing failed unexpectedly as there are not ' +
                       '9 columns, aborting.', 'critical')
            raise ValueError('sacct output does not have 9 columns. Has:' +
                             '{}: {}'.format(len(sacct[0]), sacct[0]))
        jobids = [i[0] for i in squeue]
        for sinfo in sacct:
            # Skip job steps, only index whole jobs
            if '.' in sinfo[0]:
                _logme.log('Skipping {} '.format(sinfo[0]) +
                           "in sacct processing as it is a job part.",
                           'verbose')
                continue
            # These are the values I expect
            try:
                [sid, sname, suser, spartition, sstate,
                 snodelist, snodes, scpus, scode] = sinfo
                sid, sarr = normalize_job_id(sid)
            except ValueError as err:
                _logme.log('sacct parsing failed with error {} '.format(err) +
                           'due to an incorrect number of entries.\n' +
                           'Contents of sinfo:\n{}\n'.format(sinfo) +
                           'Expected 10 values\n:' +
                           '[sid, sarr, sname, suser, spartition, sstate, ' +
                           'snodelist, snodes, scpus, scode]',
                           'critical')
                raise
            # Skip jobs that were already in squeue
            if sid in jobids:
                _logme.log('{} still in squeue output'.format(sid),
                           'verbose')
                continue
            scode = int(scode.split(':')[-1])
            squeue.append((sid, sarr, sname, suser, spartition, sstate,
                           snodelist, snodes, scpus, scode))
    else:
        _logme.log('No job info in sacct', 'debug')

    # Sanitize data
    for sinfo in squeue:
        if len(sinfo) == 10:
            [sid, sarr, sname, suser, spartition, sstate, sndlst,
             snodes, scpus, scode] = sinfo
        else:
            _sys.stderr.write('{}'.format(repr(sinfo)))
            raise _ClusterError('Queue parsing error, expected 10 items '
                                'in output of squeue and sacct, got {}\n'
                                .format(len(sinfo)))
        if partition and spartition != partition:
            continue
        if not isinstance(sid, (_str, _txt)):
            sid = str(sid) if sid else None
        else:
            sarr = None
        if not isinstance(snodes, _int):
            snodes = int(snodes) if snodes else None
        if not isinstance(scpus, _int):
            scpus = int(scpus) if snodes else None
        if not isinstance(scode, _int):
            scode = int(scode) if scode else None
        sstate = sstate.lower()
        # Convert user from ID to name
        if suser.isdigit():
            suser = _pwd.getpwuid(int(suser)).pw_name
        if user and suser != user:
            continue
        # Attempt to parse nodelist
        snodelist = []
        if sndlst:
            if nodequery.search(sndlst):
                nsplit = nodequery.findall(sndlst)
                for nrg in nsplit:
                    node, rge = nrg
                    if not rge:
                        snodelist.append(node)
                    else:
                        for reg in rge.strip('[]').split(','):
                            # Node range
                            if '-' in reg:
                                start, end = [int(i) for i in reg.split('-')]
                                for i in range(start, end):
                                    snodelist.append('{}{}'.format(node, i))
                            else:
                                snodelist.append('{}{}'.format(node, reg))
            else:
                snodelist = sndlst.split(',')

        yield (sid, sarr, sname, suser, spartition, sstate, snodelist,
               snodes, scpus, scode)


def parse_strange_options(option_dict):
    """Parse all options that cannot be handled by the regular function.

    Parameters
    ----------
    option_dict : dict
        All keyword arguments passed by the user that are not already defined
        in the Job object

    Returns
    -------
    list
        A list of strings to be added at the top of the script file
    dict
        Altered version of option_dict with all options that can't be handled
        by `fyrd.batch_systems.options.option_to_string()` removed.
    None
        Would contain additional arguments to pass to sbatch, but these are not
        needed so we just return None
    """
    outlist = []
    # Handle cores separately
    nodes = int(option_dict.pop('nodes')) if 'nodes' in option_dict else 1
    cores = int(option_dict.pop('cores')) if 'cores' in option_dict else 1

    outlist.append('#SBATCH --ntasks {}'.format(nodes))
    outlist.append('#SBATCH --cpus-per-task {}'.format(cores))

    return outlist, option_dict, None



# -*- coding: utf-8 -*-
"""
A library of useful functions used throughout the *fyrd* package.

These include functions to handle data, format outputs, handle file opening,
run commands, check file extensions, get user input, and search and format
imports.

These functions are not intended to be accessed directly and so documentation
is limited.
"""
from __future__ import with_statement
import os as _os
import re as _re
import sys as _sys
import inspect as _inspect
import argparse as _argparse
from collections import OrderedDict as _OD

import bz2
import gzip
from subprocess import Popen
from subprocess import PIPE
from time import sleep
from glob import glob as _glob

from six import text_type as _txt
from six import string_types as _str
from six import integer_types as _int
from six.moves import input as _get_input

# Progress bar handling
from tqdm import tqdm, tqdm_notebook
try:
    if str(type(get_ipython())) == "<class 'ipykernel.zmqshell.ZMQInteractiveShell'>":
        _pb = tqdm_notebook
    else:
        _pb = tqdm
except NameError:
    _pb = tqdm

from . import logme as _logme

STRPRSR = _re.compile(r'{(.*?)}')


def get_pbar(iterable, name=None, unit=None, **kwargs):
    """Return a tqdm progress bar iterable.

    If progressbar is set to False in the config, will not be shown.
    """
    from . import conf  # Avoid reciprocal import issues
    show_pb = bool(conf.get_option('queue', 'progressbar', True))
    if 'desc' in kwargs:
        dname = kwargs.pop('desc')
        name = name if name else dname
    if 'disable' in kwargs:
        disable = kwargs['disable']
    else:
        disable = False if show_pb else True
    return _pb(iterable, desc=name, unit=unit, disable=disable, **kwargs)


###############################################################################
#                               Useful Classes                                #
###############################################################################


class CustomFormatter(_argparse.ArgumentDefaultsHelpFormatter,
                      _argparse.RawDescriptionHelpFormatter):

    """Custom argparse formatting."""

    pass


class CommandError(Exception):

    """A custom exception."""

    pass


###############################################################################
#                               Misc Functions                                #
###############################################################################


def string_getter(string):
    """Parse a string for `{}`, `{#}`, and `{string}`.

    Parameters
    ----------
    string : str

    Returns
    -------
    ints : set
        A set of ints containing all `{#}` values
    vrs : set
        A set of `{string}` values

    Raises
    ------
    ValueError
        If both `{}` and `{#}` are passed
    """
    locs = STRPRSR.findall(string)
    ints = {int(i) for i in locs if i.isdigit()}
    strs = {i for i in locs if not i.isdigit()}
    if ints and '{}' in string:
        raise ValueError('Cannot parse string with both numbered and '
                         'unnumbered braces')
    return ints, strs


def parse_glob(string, get_vars=None):
    """Return a list of files that match a simple regex glob.

    Parameters
    ----------
    string : str
    get_vars : list
        A list of variable names to search for. The string must contain these
        variables in the form `{variable}`. These variables will be temporarily
        replaced with a `*` and then run through `glob.glob` to generate a list
        of files. This list is then parsed to create the output.

    Returns
    -------
    dict
        Keys are all files that match the string, values are None if `get_vars`
        is not passed. If `get_vars` is passed, the values are dictionaries
        of `{'variable': 'result'}`. e.g. for `{name}.txt` and `hi.txt`::

            {hi.txt: {name: 'hi'}}

    Raises
    ------
    ValueError
        If blank or numeric variable names are used or if get_vars returns
        multiple different names for a file.
    """
    get_vars = listify(get_vars)
    test_string = STRPRSR.sub('*', string)
    files = _glob(test_string)
    if not get_vars:
        return _OD([(f, None) for f in files])
    results = _OD([(f, {}) for f in files])

    int_vars, str_vars = string_getter(string)
    if '{}' in string or int_vars:
        raise ValueError('Cannot have numeric placeholders in file strings ',
                         "i.e. no '{0}', '{1}', '{}', etc")
    for var in get_vars:
        if var not in str_vars:
            _logme.log('Variable {0} not in search string: {1}'
                       .format(var, string), 'warn')
            continue
        # Turn search string into a regular expression
        test_var = var if var.startswith('{') else '{' + var + '}'
        test_string = STRPRSR.sub('.*?', string.replace(test_var, '(.*?)'))
        test_string = _re.sub(r'([^.])\*', r'\1.*', test_string)
        test_string = _re.sub(r'^\*', r'.*', test_string)
        # Replace terminal non-greedy operators so we parse the whole string
        if test_string.endswith('?'):
            test_string = test_string[:-1]
        if test_string.endswith('?)'):
            test_string = test_string[:-2] + ')'
        test_regex = _re.compile(test_string)
        # Add to file dict
        for fl in files:
            vrs = test_regex.findall(fl)
            ulen = len(set(vrs))
            if ulen != 1:
                _logme.log('File {0} has multiple values for {1}: {2}'
                           .format(fl, test_var, vrs), 'critical')
                raise ValueError('Invalid file search string')
            if ulen == 0:
                _logme.log('File {0} has no results for {1}'
                           .format(fl, test_var), 'error')
                continue
            results[fl][var] = vrs[0]
    return results


def file_getter(file_strings, variables, extra_vars=None, max_count=None):
    """Get a list of files and variable values using the search string.

    The file strings can contain standard unix glob (like `*`) and variable
    containing strings in the form `{name}`.

    For example, a file_string of `{dir}/*.txt` will match every file that
    ends in `.txt` in every directory relative to the current path.

    The result for a directory name test with two files named 1.txt and 2.txt
    is a list of::

        [(('dir/1.txt'), {'dir': 'test'}),
         (('dir/2.txt'), {'dir': 'test'})]

    This is repeated for every file_string in file_strings, and the following
    tests are done:

        1. All file_strings must result in identical numbers of files
        2. All variables must have only a single value in every file string

    If there are multiple file_strings, they are added to the result x in
    order, but the dictionary remains the same as variables must be shared. If
    multiple file_strings are provided the results are combined by alphabetical
    order.

    Parameters
    ----------
    file_strings : list of str
        List of search strings, e.g. `*/*`, `*/*.txt`, `{dir}/*.txt` or
        `{dir}/{file}.txt`
    variables : list of str
        List of variables to look for
    extra_vars : list of str, optional
        A list of additional variables specified in a very precise format::

            new_var:orig_var:regex:sub_str

            or

            new_var:value

        The orig_var must correspond to a variable in variables. var will be
        generated by running re.sub(regex, sub_str, string) where string is
        the result of orig_var for the given file set
    max_count : int, optional
        Max number of file_strings to parse, default is all.

    Returns
    -------
    list
        A list of files. Each list item will be a two-item tuple of
        `(files, variables)`. Files will be a tuple with the same length as
        max_count, or file_strings if max_count is None. Variables will be
        a dictionary of all variables and extra_vars for this file set. e.g.::

            [((file1, dir1, file2), {var1: val, var2: val})]

    Raises
    ------
    ValueError
        Raised if any of the above tests are not met.
    """
    # Make extra_var an empty list if None
    extra_vars = listify(extra_vars) if extra_vars else []

    # Get all file information
    files = []
    count = 0
    for file_string in file_strings:
        files.append(parse_glob(file_string, variables))
        count += 1
        if max_count and count == max_count:
            break

    # Make sure all files have values for variables
    var_vals = {i: [] for i in variables}
    empty = {}
    for f in files:
        for pvars in f.values():
            # It's fine if the file string has no variables
            if not pvars:
                continue
            for var, val in pvars.items():
                if not val:
                    if var in empty:
                        empty[var].append(f)
                    else:
                        empty[var] = [f]
                else:
                    var_vals[var].append(val)
    fail = False
    if empty:
        _logme.log('The following variables had no '
                   'result in some files, cannot continue:\n'
                   '\n'.join(
                       ['{0} files: {1}'.format(i, j) for i, j in empty.items()]
                   ), 'critical')
        fail = True

    # Join files
    bad = []

    results = []
    for file_info in zip(*[fl.items() for fl in files]):
        good = True
        # Make sure dicts are compatible and make combined dict
        final_dict = {}
        final_files = tuple([_os.path.abspath(fl[0]) for fl in file_info])
        bad_dcts = []
        for dct in [f[1] for f in file_info]:
            if not dct:
                continue
            final_dict.update(dct)
            bad_dcts.append(dct)
            for var, val in dct.items():
                for fl in file_info:
                    if var in fl[1] and val != fl[1][var]:
                        good = False
        if not good:
            bad.append((final_files, bad_dcts))
            break
        for extra_var in extra_vars:
            try:
                evari = extra_var.split(':')
                if len(evari) == 2:
                    final_dict[evari[1]] = evari[2]
                    continue
                var, orig_var, parse_str, sub_str = evari
            except ValueError:
                _logme.log(
                    '{} is malformatted should be: '.format(extra_var) +
                    'either new_var:orig_var:regex:sub '
                    'or variable:value', 'critical'
                )
                raise
            if orig_var not in final_dict:
                raise ValueError(
                    'Extra variable {0} sets {1} as '.format(var, orig_var) +
                    'the original variable, but it is not in the dict for '
                    '{0}'.format(final_files)
                )
            final_dict[var] = _re.sub(parse_str, sub_str, final_dict[orig_var])
        results.append((final_files, final_dict))

    # Take care of bad items
    if empty:
        _logme.log('The following file combinations had mismatched variables, '
                   'cannot continue:\n'
                   '\n'.join(
                       ['{0} dicts: {1}'.format(i, j) for i, j in bad]
                   ), 'critical')
        fail = True

    if fail:
        raise ValueError('File parsing failure')

    return results


def listify(iterable):
    """Try to force any iterable into a list sensibly."""
    if isinstance(iterable, list):
        return iterable
    if isinstance(iterable, (_str, _txt, _int, float)):
        return [iterable]
    if not iterable:
        return []
    try:
        iterable = list(iterable)
    except TypeError:
        iterable = [iterable]
    return iterable


def merge_lists(lists):
    """Turn a list of lists into a single list."""
    outlist = []
    for lst in listify(lists):
        outlist += lst
    return outlist


def write_iterable(iterable, outfile):
    """Write all elements of iterable to outfile."""
    with open_zipped(outfile, 'w') as fout:
        fout.write('\n'.join(iterable))


def indent(string, prefix='    '):
    """Replicate python3's textwrap.indent for python2.

    Parameters
    ----------
    string : str
        Any string.
    prefix : str
        What to indent with.

    Returns
    -------
    str
        Indented string
    """
    out = ''
    for i in string.split('\n'):
        out += '{}{}\n'.format(prefix, i)
    return out


def is_exc(x):
    """Check if x is the output of sys.exc_info().

    Returns
    -------
    bool
        True if matched the output of sys.exc_info().
    """
    return bool(isinstance(x, tuple)
                and len(x) == 3
                and issubclass(BaseException, x[0]))


###############################################################################
#                               File Management                               #
###############################################################################


def open_zipped(infile, mode='r'):
    """Open a regular, gzipped, or bz2 file.

    If infile is a file handle or text device, it is returned without
    changes.

    Returns
    -------
    text mode file handle.
    """
    mode   = mode[0] + 't'
    if hasattr(infile, 'write'):
        return infile
    if isinstance(infile, _str):
        if infile.endswith('.gz'):
            return gzip.open(infile, mode)
        if infile.endswith('.bz2'):
            if hasattr(bz2, 'open'):
                return bz2.open(infile, mode)
            else:
                return bz2.BZ2File(infile, mode)
        return open(infile, mode)


def exp_file(infile):
    """Return an expanded path to a file."""
    return _os.path.expandvars(
        _re.sub(
            '~', '$HOME', infile
        )
    )


def cmd_or_file(string):
    """If string is a file, return the contents, else return the string.

    Parameters
    ----------
    string : str
        Path to a file or any other string

    Returns
    -------
    script : str
        Either the contents of the file if string is a file or just the
        contents of string.
    """
    if _os.path.isfile(string):
        with open_zipped(string) as fin:
            command = fin.read().strip()
    else:
        command = string.strip()
    return command


def block_read(files, size=65536):
    """Iterate through a file by blocks."""
    while True:
        b = files.read(size)
        if not b:
            break
        yield b


def count_lines(infile, force_blocks=False):
    """Return the line count of a file as quickly as possible.

    Uses `wc` if avaialable, otherwise does a rapid read.
    """
    if which('wc') and not force_blocks:
        _logme.log('Using wc', 'debug')
        if infile.endswith('.gz'):
            cat = 'zcat'
        elif infile.endswith('.bz2'):
            cat = 'bzcat'
        else:
            cat = 'cat'
        command = "{cat} {infile} | wc -l | awk '{{print $1}}'".format(
            cat=cat, infile=infile
        )
        return int(cmd(command)[1])
    else:
        _logme.log('Using block read', 'debug')
        with open_zipped(infile) as fin:
            return sum(bl.count("\n") for bl in block_read(fin))


def split_file(infile, parts, outpath='', keep_header=False):
    """Split a file in parts and return a list of paths.

    .. note:: Linux specific (uses wc).

    If has_header is True, the top line is stripped off the infile prior to
    splitting and assumed to be the header.

    Parameters
    ----------
    outpath : str, optional
        The directory to save the split files.
    keep_header : bool, optional
        Add the header line to the top of every file.

    Returns
    -------
    list
        Paths to split files.
    """
    # Determine how many reads will be in each split sam file.
    _logme.log('Getting line count', 'debug')

    num_lines = int(count_lines(infile)/int(parts)) + 1

    # Subset the file into X number of jobs, maintain extension
    cnt       = 0
    currjob   = 1
    suffix    = '.split_' + str(currjob).zfill(4) + '.' + infile.split('.')[-1]
    file_name = _os.path.basename(infile)
    run_file  = _os.path.join(outpath, file_name + suffix)
    outfiles  = [run_file]

    # Actually split the file
    _logme.log('Splitting file', 'debug')
    with open_zipped(infile) as fin:
        header = fin.readline() if keep_header else ''
        sfile = open_zipped(run_file, 'w')
        sfile.write(header)
        for line in fin:
            cnt += 1
            if cnt < num_lines:
                sfile.write(line)
            elif cnt == num_lines:
                sfile.write(line)
                sfile.close()
                currjob += 1
                suffix = '.split_' + str(currjob).zfill(4) + '.' + \
                    infile.split('.')[-1]
                run_file = _os.path.join(outpath, file_name + suffix)
                sfile = open_zipped(run_file, 'w')
                outfiles.append(run_file)
                sfile.write(header)
                cnt = 0
        sfile.close()
    _logme.log('Split files: {}'.format(outfiles), 'debug')
    return tuple(outfiles)


def is_exe(fpath):
    """Return True is fpath is executable."""
    return _os.path.isfile(fpath) and _os.access(fpath, _os.X_OK)


def file_type(infile):
    """Return file type after stripping gz or bz2."""
    name_parts = infile.split('.')
    if name_parts[-1] == 'gz' or name_parts[-1] == 'bz2':
        name_parts.pop()
    return name_parts[-1]


def is_file_type(infile, types):
    """Return True if infile is one of types.

    Parameters
    ----------
    infile : str
        Any file name
    types : list
        String or list/tuple of strings (e.g `['bed', 'gtf']`)

    Returns
    -------
    is_file_type : bool
    """
    if hasattr(infile, 'write'):
        infile = infile.name
    types = listify(types)
    for typ in types:
        if file_type(infile) == typ:
            return True
    return False


###############################################################################
#                              Running Commands                               #
###############################################################################


def cmd(command, args=None, stdout=None, stderr=None, tries=1):
    """Run command and return status, output, stderr.

    Parameters
    ----------
    command : str
        Path to executable.
    args : tuple, optional
        Tuple of arguments.
    stdout : str, optional
        File or open file like object to write STDOUT to.
    stderr : str, optional
        File or open file like object to write STDERR to.
    tries : int, optional
        Number of times to try to execute. 1+

    Returns
    -------
    exit_code : int
    STDOUT : str
    STDERR : str
    """
    tries = int(tries)
    assert tries > 0
    count = 1
    if isinstance(command, (list, tuple)):
        if args:
            raise ValueError('Cannot submit list/tuple command as ' +
                             'well as args argument')
        command = ' '.join(command)
    assert isinstance(command, _str)
    if args:
        if isinstance(args, (list, tuple)):
            args = ' '.join(args)
        args = command + args
    else:
        args = command
    _logme.log('Running {} as {}'.format(command, args), 'verbose')
    while True:
        try:
            pp = Popen(args, shell=True, universal_newlines=True,
                       stdout=PIPE, stderr=PIPE)
        except FileNotFoundError:
            _logme.log('{} does not exist'.format(command), 'critical')
            raise
        out, err = pp.communicate()
        code = pp.returncode
        if code == 0 or count == tries:
            break
        _logme.log('Command {} failed with code {}, retrying.'
                   .format(command, code), 'warn')
        sleep(1)
        count += 1
    _logme.log('{} completed with code {}'.format(command, code), 'debug')
    if stdout:
        with open_zipped(stdout, 'w') as fout:
            fout.write(out)
    if stderr:
        with open_zipped(stderr, 'w') as fout:
            fout.write(err)
    return code, out.rstrip(), err.rstrip()


def export_run(function, args, kwargs):
    """Execute a function after first exporting all imports."""
    kwargs['imports'] = export_imports(function, kwargs)
    return function(*args, **kwargs)


def which(program):
    """Replicate the UNIX which command.

    Taken verbatim from:
        stackoverflow.com/questions/377017/test-if-executable-exists-in-python

    Parameters
    ----------
    program : str
        Name of executable to test.

    Returns
    -------
    str or None
        Path to the program or None on failure.
    """
    fpath, program = _os.path.split(program)
    if fpath:
        if is_exe(program):
            return _os.path.abspath(program)
    else:
        for path in _os.environ["PATH"].split(_os.pathsep):
            path = path.strip('"')
            exe_file = _os.path.join(path, program)
            if is_exe(exe_file):
                return _os.path.abspath(exe_file)

    return None


def check_pid(pid):
    """Check For the existence of a unix pid."""
    try:
        _os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


###############################################################################
#                       Option and Argument Management                        #
###############################################################################


def replace_argument(args, find_string, replace_string, error=True):
    """Replace find_string with replace string in a tuple or dict.

    If dict, the values are replaced, not the keys.

    Note: args can also be a list, in which case the first item is assumed
    to be a tuple, and the second a dictionary

    Parameters
    ----------
    args : list/tuple/dict
        Tuple or dict of args
    find_string : str
        A string to search for
    replace_string : str
        A string to replace with
    error : bool
        Raise ValueError if replacement fails

    Returns
    -------
    The same object as was passed, with alterations made.
    """
    double = False
    if isinstance(args, list):
        args, kwargs = args
        double = True
    elif isinstance(args, tuple):
        kwargs = None
    elif isinstance(args, dict):
        kwargs = args.copy()
        args   = None
    else:
        raise ValueError('args must be list/tuple/dict, is {}\nval: {}'
                         .format(type(args), args))

    if not args and not kwargs:
        msg = 'No arguments or keyword arguments found'
        if error:
            raise ValueError(msg)
        else:
            _logme.log(msg, 'warn')
            if double:
                return None, None
            else:
                return None

    found = False
    newargs = tuple()
    if args:
        for arg in listify(args):
            if isinstance(arg, _str) and find_string in arg:
                arg = arg.format(**{find_string.strip('{}'): replace_string})
                found = True
            newargs += (arg,)
    newkwds = {}
    if kwargs:
        for arg, value in kwargs.items():
            if isinstance(value, _str) and find_string in value:
                value = replace_string
                found = True
            newkwds[arg] = value

    if found is not True:
        msg = 'Could not find {}'.format(find_string)
        if error:
            raise ValueError(msg)
        else:
            _logme.log(msg, 'warn')
            if double:
                return None, None
            else:
                return None

    if double:
        return [newargs, newkwds]
    else:
        if newargs:
            return newargs
        else:
            return newkwds


def opt_split(opt, split_on):
    """Split options by chars in split_on, merge all into single list.

    Parameters
    ----------
    opt : list
        A list of strings, can be a single string.
    split_on : list
        A list of characters to use to split the options.

    Returns
    -------
    list
        A single merged list of split options, uniqueness guaranteed, order
        not.
    """
    opt = listify(opt)
    split_on = listify(split_on)
    final_list = []
    for o in opt:
        final_list += _re.split('[{}]'.format(''.join(split_on)), o)
    return list(set(final_list)) # Return unique options only, order lost.


###############################################################################
#                                 User Input                                  #
###############################################################################


def get_yesno(message, default=None):
    """Get yes/no answer from user.

    Parameters
    ----------
    message : str
        A message to print, an additional space will be added.
    default : {'y', 'n'}, optional
        One of `{'y', 'n'}`, the default if the user gives no answer. If None,
        answer forced.

    Returns
    -------
    bool
        True on yes, False on no
    """
    if default:
        if default.lower().startswith('y'):
            tailstr = '[Y/n] '
        elif default.lower().startswith('n'):
            tailstr = '[y/N] '
        else:
            raise ValueError('Invalid default')
    else:
        tailstr = '[y/n] '
    message = message + tailstr if message.endswith(' ') \
                else message + ' ' + tailstr
    ans = get_input(message, 'yesno', default)
    if ans.lower().startswith('y'):
        return True
    elif ans.lower().startswith('n'):
        return False
    else:
        raise ValueError('Invalid response: {}'.format(ans))


def get_input(message, valid_answers=None, default=None):
    """Get input from the command line and check answers.

    Allows input to work with python 2/3

    Parameters
    ----------
    message : str
        A message to print, an additional space will be added.
    valid_answers : list
        A list of answers to accept, if None, ignored.  Case insensitive. There
        is one special option here: 'yesno', this allows all case insensitive
        variations of y/n/yes/no.
    default : str
        The default answer.

    Returns
    -------
    response : str
    """
    if not message.endswith(' '):
        message = message + ' '
    if valid_answers:
        if isinstance(valid_answers, _str):
            if valid_answers.lower() == 'yesno':
                valid_answers = ['yes', 'no', 'y', 'n']
            else:
                valid_answers = [valid_answers]
        if not isinstance(valid_answers, (list, tuple, set, frozenset)):
            _logme.log('valid_answers must be a list, is {}'
                       .format(type(valid_answers)), 'critical')
            raise ValueError('Invalid argument')
        valid_answers = [i.lower() for i in valid_answers]
        while True:
            ans = _get_input(message)
            if not ans and default:
                return default
            if ans.lower() in valid_answers:
                return ans
            else:
                _logme.log('Invalid response to input question', 'debug')
                _sys.stderr.write('Invalid response: {}\n'.format(ans) +
                                  'Valid responses: {}\n'
                                  .format(valid_answers) +
                                  'Please try again.\n')
    else:
        return _get_input(message)


###############################################################################
#                                   Imports                                   #
###############################################################################


def syspath_fmt(syspaths):
    """Take a list of paths and return a sys of sys.path.append strings."""
    outlist = []
    for pth in listify(syspaths):
        if 'sys.path' in pth:
            outlist.append(pth)
            continue
        if _os.path.exists(pth):
            outlist.append("sys.path.append('{}')".format(
                _os.path.abspath(pth)
            ))
        else:
            raise OSError('Paths must exist, {} does not.'
                          .format(pth))
    return '\n'.join(outlist)


PROT_IMPT = """\
try:
    {}
except ImportError:
    pass
"""


def normalize_imports(imports, prot=True):
    """Take a heterogenous list of imports and normalize it.

    Parameters
    ----------
    imports : list
        A list of strings, formatted differently.
    prot : bool
        Protect imports with try..except blocks

    Returns
    -------
    list
        A list of strings that can be used for imports
    """
    out_impts  = []
    prot_impts = []
    path_impts = []
    imports = listify(imports)
    if not imports:
        return []
    for imp in imports:
        if not isinstance(imp, _str):
            raise ValueError('All imports must be strings')
        if imp.startswith('try:'):
            prot_impts.append(imp.rstrip())
        elif imp.startswith('import') or imp.startswith('from'):
            out_impts.append(imp.rstrip())
        elif imp.startswith('sys.path.append')\
            or imp.startswith('sys.path.insert'):
            path_impts.append(imp.rstrip())
        else:
            if imp.startswith('@'):
                continue
            out_impts.append('import {}'.format(imp))

    if prot:
        for imp in out_impts:
            prot_impts.append(PROT_IMPT.format(imp))
        out = prot_impts
    else:
        out = out_impts + prot_impts

    # Remove duplicates
    out = list(set(out))

    # Add PATHS
    if path_impts:
        out = list(set(path_impts)) + out

    return out


def get_function_path(function):
    """Return path to module defining a function if it exists."""
    mod = _inspect.getmodule(function)
    if mod and mod != '__main__':
        return _os.path.dirname(_inspect.getabsfile(function))
    else:
        return None


def update_syspaths(function, kwds=None):
    """Add function path to 'syspaths' in kwds."""
    if kwds:
        syspaths = listify(kwds['syspaths']) if 'syspaths' in kwds else []
    else:
        syspaths = []
    return [get_function_path(function)] + syspaths


def import_function(function, mode='string'):
    """Return an import string for the function.

    Attempts to resolve the parent module also, if the parent module is a file,
    ie it isn't __main__, the import string will include a call to
    sys.path.append to ensure the module is importable.

    If this function isn't defined by a module, returns an empty string.

    Parameters
    ----------
    mode : {'string', 'list'}, optional
        string/list, return as a unified string or a list.
    """
    if not callable(function):
        raise ValueError('Function must be callable, {} is not'
                         .format(function))
    if mode not in ['string', 'list']:
        raise ValueError("Invalid mode {}, must be 'list' or 'string'"
                         .format(mode))

    if _inspect.ismethod(function):
        name = (dict(_inspect.getmembers(function.__self__))['__class__']
                .__name__)
    else:
        name    = function.__name__

    # Attempt to resolve defining file
    parent = _inspect.getmodule(function)

    imports = []
    if parent and parent.__name__ != '__main__':
        path   = _os.path.dirname(parent.__file__)
        module = parent.__name__
        # If module is the child of a package, change the directory up to the
        # parent
        if '.' in module:
            path = _os.path.abspath(
                _os.path.join(
                    path, *['..' for i in range(module.count('.'))]
                )
            )
        imports.append("sys.path.append('{}')".format(path))
        imports.append('import {}'.format(module))
        imports.append('from {} import *'.format(module))
        imports.append('from {} import {}'.format(module, name))

    return imports if mode == 'list' else '\n'.join(imports)


def get_imports(function, mode='string'):
    """Build a list of potentially useful imports from a function handle.

    Gets:

    - All modules from globals()
    - All modules from the function's globals()
    - All functions from the function's globals()

    Modes:

    string:
        Return a list of strings formatted as unprotected import calls
    prot:
        Similar to string, but with try..except blocks
    list:
        Return two lists: (import name, module name) for modules and (import
        name, function name, module name) for functions

    Parameters
    ----------
    function : callable
        A function handle
    mode : str
        A string corresponding to one of the above modes

    Returns
    -------
    str or list
    """
    if mode not in ['string', 'prot', 'list']:
        raise ValueError('mode must be one of string/prot/list')

    rootmod  = _inspect.getmodule(function)

    imports      = []
    func_imports = []

    # For interactive sessions
    members = dict(_inspect.getmembers(function))
    locations = [members]
    if '__globals__' in members:
        locations.append(members['__globals__'])
    for location in locations:
        for name, item in location.items():
            if name.startswith('__'):
                continue
            # Modules
            if _inspect.ismodule(item):
                imports.append((name, item.__name__))
            # Functions
            elif callable(item):
                try:
                    func_imports.append((name, item.__name__, item.__module__))
                except AttributeError:
                    pass

    # Import all modules in the root module
    imports += [(k,v.__name__)
                for k,v in _inspect.getmembers(rootmod, _inspect.ismodule)
                if not k.startswith('__')]

    # Make unique
    imports      = sorted(list(set(imports)), key=_sort_imports)
    func_imports = sorted(list(set(func_imports)), key=_sort_imports)

    _logme.log('Imports: {}'.format(imports), 'debug')
    _logme.log('Func imports: {}'.format(func_imports), 'debug')

    # Create a sane set of imports
    ignore_list = ['os', 'sys', 'dill', 'pickle', '__main__']
    filtered_imports      = []
    filtered_func_imports = []
    for iname, name in imports:
        if iname in ignore_list:
            continue
        if name.startswith('@') or iname.startswith('@'):
            continue
        filtered_imports.append((iname, name))
    for iname, name, mod in func_imports:
        if iname in ignore_list:
            continue
        if name.startswith('@') or iname.startswith('@'):
            continue
        filtered_func_imports.append((iname, name, mod))

    if mode == 'list':
        return filtered_imports, filtered_func_imports

    import_strings = []
    for iname, name in filtered_imports:
        names = name.split('.')
        if names[0] == '__main__':
            continue
        if iname != name:
            if len(names) > 1:
                if '.'.join(names[1:]) != iname:
                    import_strings.append(
                        'from {} import {} as {}'
                        .format('.'.join(names[:-1]), names[-1], iname)
                    )
                else:
                    import_strings.append(
                        'from {} import {}'
                        .format(names[0], '.'.join(names[1:]))
                    )
            else:
                import_strings.append(
                    ('import {} as {}').format(name, iname)
                )
        else:
            import_strings.append('import {}'.format(name))

    # Function imports
    for iname, name, mod in filtered_func_imports:
        if mod == '__main__':
            continue
        if iname == name:
            import_strings.append('from {} import {}'.format(mod, name))
        else:
            import_strings.append('from {} import {} as {}'
                                  .format(mod, name, iname)
                                 )

    if mode == 'string':
        return import_strings

    elif mode == 'prot':
        return normalize_imports(import_strings, prot=True)

    else:
        raise ValueError('Mode changed unexpectedly')


def export_globals(function):
    """Add a function's globals to the current globals."""
    rootmod = _inspect.getmodule(function)
    globals()[rootmod.__name__] = rootmod
    for k, v in _inspect.getmembers(rootmod, _inspect.ismodule):
        if not k.startswith('__'):
            globals()[k] = v


def get_all_imports(function, kwds, prot=False):
    """Get all imports from a function and from kwds.

    Parameters
    ----------
    function : callable
        A function handle
    kwds : dict
        A dictionary of keyword arguments
    prot : bool
        Wrap all import in try statement

    Returns
    -------
    list
        Imports
    """
    imports  = listify(kwds['imports'] if 'imports' in kwds else None)
    imports  = normalize_imports(imports, prot=False)
    imports += get_imports(function, mode='string')
    return normalize_imports(imports, prot=prot)


def export_imports(function, kwds):
    """Get imports from a function and from kwds.

    Also sets globals and adds path to module to sys path.

    Parameters
    ----------
    function : callable
        A function handle
    kwds : dict
        A dictionary of keyword arguments

    Returns
    -------
    list
        imports + sys.path.append for module path
    """
    export_globals(function)
    return import_function(function, 'list') + get_all_imports(function, kwds)


def _sort_imports(x):
    """Sort a list of tuples and strings, for use with sorted."""
    if isinstance(x, tuple):
        if x[1] == '__main__':
            return 0
        return x[1]
    return x



# -*- coding: utf-8 -*-
"""
Get and set config file options.

The functions defined here provide an easy way to access the config file
defined by CONFIG_FILE (default ~/.fyrd/config.txt) and the config.get('jobs',
'profile_file') (default ~/.fyrd/profile.txt).

Both files are managed by Python's ConfigParser class.

To work with profiles, use the `get_profile()` and `set_profile()` functions.
Note that all options must be allowed in the `options.py` file before they can
be added to a profile.

Options will also be pre-sanitized before being added to profile. e.g. 'mem':
    '2GB' will become 'mem': 2000.

Functions
---------
get_option: Get a single key or section.
set_option: Set a single key or section.
set_option: Set a single key or section.
delete: Delete a config item.
create_config: Create an initial config from a dictionary and the defaults
create_config_interactive: Interact with the user to create a new config

Examples
--------
>>> import fyrd
>>> i = fyrd.conf.get_option('queue', 'sleep_len')
>>> c = fyrd.conf.set_option('queue', 'sleep_len', 0.125)
>>> fyrd.conf.get_option('queue', 'sleep_len')
0.125

"""
from __future__ import print_function
import re
import os       as _os
import readline as _rl
from textwrap import dedent as _dnt
try:
    import configparser as _configparser
except ImportError:
    import ConfigParser as _configparser

import six

from . import run     as _run
from . import logme   as _logme
from . import batch_systems as _batch
_opt = _batch.options


###############################################################################
#                            Configurable Defaults                            #
###############################################################################

"""
Where configuration files will be kept
"""
CONFIG_PATH  = _os.path.join(_os.environ['HOME'], '.fyrd')

"""
Where the main config will be kept.
"""
CONFIG_FILE  = _os.path.join(CONFIG_PATH, 'config.txt')

# Set default options
DEFAULTS = {
    'queue': {
        'max_jobs':         1000,
        'sleep_len':        1,
        'queue_update':     2,
        'res_time':         2700, # Time to wait if job is postponed before fail
        'queue_type':       'auto',
        'sbatch':           None, # Path to sbatch command
        'qsub':             None, # Path to qsub command
        'progressbar':      True,
        'queue_maximums':   _os.path.join(
            CONFIG_PATH, 'queue_maximums.json'
        ),
    },
    'jobs': {
        'clean_files':     True,
        'clean_outputs':   False,
        'file_block_time': 30,
        'scriptpath':      None,
        'outpath':         None,
        'suffix':          'cluster',
        'auto_submit':     True,
        'generic_python':  False,
        'profile_file':    _os.path.join(
            CONFIG_PATH, 'profiles.txt'
        )
    },
    'notify': {
        'mode': 'linux',  # Or SMTP
        'wait_time': 240, # time in seconds
        'notify_address': None,
        'smtp_host': 'smtp.gmail.com',
        'smtp_port': 587,
        'smtp_tls': True,
        'smtp_from': None,
        'smtp_user': None,
        'smtp_passfile': None
    },
    'local': {
        'server_uri':      None,
        'max_jobs':        None,
        'local_clean_days': 7,
    }
}

CONF_HELP = {
    'summary': _dnt(
        """
        The following options and sections are recognized and defined by the
        DEFAULTS dictionary in the config.py file. The can be updated in the
        config file.

        Any options added to the config file not present here will be ignored
        or deleted, any options added which do not match the types below will
        be overwritten with the defaults.
        """
    ),
    'queue': _dnt(
        """
        [queue]
        Define options for queue handling

        Options
        -------
        max_jobs : int
            sets the maximum number of running jobs before submission will
            pause and wait for the queue to empty
        sleep_len : int
            sets the amount of time the program will wait between submission
            attempts
        queue_update : int
            sets the amount of time between refreshes of the queue.
        res_time : int
            Time in seconds to wait if a job is in an uncertain state, usually
            preempted or suspended. These jobs often resolve into running or
            completed again after some time so it makes sense to wait a bit,
            but not forever.  The default is 45 minutes: 2700 seconds.
        queue_type : str
            the type of queue to use, one of the batch systems (e.g. 'slurm')
            or 'auto'. Default is auto to auto-detect the queue.
        sbatch : str
            A path to the sbatch executable, only required for slurm mode if
            sbatch is not in the PATH.
        qsub : str
            A path to the qsub executable, only required for torque mode if
            sbatch is not in the PATH.
        progressbar : bool
            Show a progress bar when waiting for jobs
        queue_maximums : str
            Path to a JSON file that sets the queue maximums, the file must
            be in the format {'<queue>': {'max_<keyword>': val}}, where
            <queue> is a batch system partition/queue and <keyword> is a valid
            keyword argument (e.g. cores, mem, time, etc)
        """
    ),
    'jobs': _dnt(
        """
        [jobs]
        Set the options for managing job submission and getting

        Options
        -------
        clean_files : bool
            means that by default files will be deleted when job completes
        clean_outputs : bool
            is the same but for output files (they are saved first)
        file_block_time : int
            Max amount of time to block after job completes in the queue while
            waiting for output files to appear.  Some queues can take a long
            time to copy files under load, so it is worth setting this high, it
            won't block unless the files do not appear.
        scriptpath : str
            Path to write all script files by default, must be globally cluster
            accessible. Note: this is *not* the runtime path, just where files
            are written to.
        outpath : str
            Path to write all output files to by by default, must be globally
            cluster accessible.
        suffix : str
            The suffix to use when writing scripts and output files
        auto_submit : bool
            If wait() or get() are called prior to submission, auto-submit the
            job. Otherwise throws an error and returns None
        generic_python : bool
            Use /usr/bin/env python instead of the current executable, not
            advised, but sometimes necessary.
        profile_file : str
            the config file where profiles are defined.
        """
    ),
    'notify': _dnt(
        """
        [notify]
        User notification option, used mostly to email completion info.

        Options
        -------
        mode : 'linux', 'smtp', or None
            If mode is linux, all other entries optional and `mail` is used.
            If mode is smtp, all other entries required, `smtp_passfile` must
            point to a read only file accessible only to the user that
            contains the password for smtp.
        wait_time : int
            Time to wait before auto-email, in seconds.
        notify_address : str
            The address to send an email to. Can also be specified at runtime
        smtp_host : str
            e.g. smtp.gmail.com
        smtp_port : int
            e.g. 587
        smtp_tls : bool
            Use TLS?
        smtp_from : str
            Server email address, e.g. a dedicated gmail adrress
        smtp_user : str, optional
            If not provided, smtp_from used
        smtp_passfile : str
            Path to a plain text, read-only, user accessible password file with
            the SMTP password. Perissions should be 400.
        """
    ),
    'local': _dnt(
        """
        [local]
        Set options just for the local queue batch system.

        Options
        -------
        server_uri : str, optional
            Hardcode the server_uri for the local queue server, this allows
            you to set the server as a remote machine if you want, but you
            must have the remote drives mounted in that case, these mounted
            drives need only be in the 'scriptpath' and 'outpath' directories.
        max_jobs : int, optional
            Set the maximum locally running jobs, defaults to all available
            cores minus 1.
        local_clean_days : int, optional
            The number of days to keep jobs in the local queue. Any jobs older
            than this will be purged from the database
        """
    )
}

# Pre-defined profiles, 'DEFAULT' is required.
DEFAULT_PROFILES = {
    'DEFAULT':     {'nodes': 1,
                    'cores': 1,
                    'mem':   4000,
                    'time':  '04:00:00'},
    'large':       {'nodes': 1,
                    'cores': 16,
                    'mem':   32000,
                    'time':  '24:00:00'},
    'small':       {'nodes': 1,
                    'cores': 1,
                    'mem':   1000,
                    'time':  '01:00:00'},
    'long':        {'nodes': 1,
                    'cores': 1,
                    'mem':   4000,
                    'time':  '96:00:00'},
    'small_clean': {'nodes': 1,
                    'cores': 1,
                    'mem':   1000,
                    'time':  '01:00:00',
                    'clean_files':   True,
                    'clean_outputs': True},
}
"""
This defines the default profiles that will be stored in the profile file.
It is intended mostly as an example and should be edited in the profile file
to customize the profiles for each cluster.

The only required profile is 'DEFAULT', it must be set and is the fallback
profile if no profile is requested on job creation.
"""


###############################################################################
#                         Do Not Edit Below This Point                        #
###############################################################################

# Create the global config objects
config   = _configparser.ConfigParser(allow_no_value=True)
"""
This is the globally accessible ConfigParser object for the config.txt file.
"""
profiles = _configparser.ConfigParser(defaults=DEFAULT_PROFILES['DEFAULT'],
                                      allow_no_value=True)
"""
This is the globally accessible ConfigParser object for handling profiles.
"""

__all__ = ['set_option', 'get_option', 'delete', 'create_config',
           'create_config_interactive', 'set_profile',
           'get_profile', 'del_profile', 'create_profiles']


###############################################################################
#                        Config Manipulation Functions                        #
###############################################################################


def get_executable(section, key, default):
    """Get an executable (e.g. qsub) from the conf if it exists.

    Parameters
    ----------
    section : str
        The config section to use (e.g. queue), if None, all sections returned.
    key : str
        The config key to get (e.g. 'max_jobs'), if None, whole section
        returned.
    default
        If the key does not exist, create it with this default value.

    Returns
    -------
    str or None
        Path to executable if found, else None
    """
    exe = get_option(section, key, default)
    if not _os.path.dirname(exe):
        exe = _run.which(exe)
    if _run.is_exe(exe):
        return exe
    return None


def get_option(section=None, key=None, default=None):
    """Get a single key or section.

    All args are optional, if they are missing, the parent section or entire
    config will be returned.

    Parameters
    ----------
    section : str
        The config section to use (e.g. queue), if None, all sections returned.
    key : str
        The config key to get (e.g. 'max_jobs'), if None, whole section
        returned.
    default
        If the key does not exist, create it with this default value.

    Returns
    -------
    option_value
        Option value if key exists, None if no key exists.

    See Also
    --------
    set_option : Set an option
    get_config : Get the entire config
    """
    load_config()

    if not section:
        out = _config_to_dict(config)

    elif section and section not in _sections(config):
        _logme.log('{} not in the config file'.format(section), 'error')
        if section in DEFAULTS:
            _logme.log('CreAting {} from DEFAULTS'.format(section), 'info')
            config.add_section(section)
            _config_from_dict(config, DEFAULTS[section], section=section)
            write_config()
            out = get_option(section, key, default)
        else:
            raise ValueError('Section not in the config file or DEFAULTS')

    elif key:
        sect = _section_to_dict(config.items(section))
        if key in sect:
            out = sect[key]
        else:
            _logme.log('{} not in the {} section of the config file'
                       .format(key, section), 'warn')
            if default:
                _logme.log('Creating new config entry {}:{} with val {}'
                           .format(section, key, default))
                set_option(section, key, default)
                out = get_option(section, key)
            elif key in DEFAULTS[section]:
                _logme.log('Creating new config entry {}:{} with val {}'
                           .format(section, key, DEFAULTS[section][key]))
                set_option(section, key, DEFAULTS[section][key])
                out = get_option(section, key)
            else:
                raise ValueError('Option not in the config file')

    elif section:
        _logme.log('Getting the whole section: {}'.format(section), 'debug')
        out = _section_to_dict(config.items(section))

    else:
        _logme.log('No option specified, returning config dictionary', 'debug')
        out = _config_to_dict(config)

    return _typecast_items(out)


def set_option(section, key, value):
    """Write a config key to the config file.

    Parameters
    ----------
    section : str
        Section of the config file to use.
    key : str
        Key to add.
    value
        Value to add for key.

    Returns
    -------
    ConfigParser
    """
    # Sanitize arguments
    section = str(section)
    key     = str(key)
    value   = str(value)

    # Edit the globals in this file
    load_config()

    if not config.has_section(section):
        _logme.log('The {} section is not in the config file and cannot be'
                   .format(section) + 'and cannot be created', 'warn')
        return None

    config.set(section, key, value)

    write_config()

    return config


def delete(section, key):
    """Delete a config item.

    Parameters
    ----------
    section : str
        Section of config file.
    key : str
        Key to delete

    Returns
    -------
    ConfigParger
    """
    load_config()

    config.remove_option(section, key)

    write_config()

    return config


def load_config():
    """Load config from the config file.

    If any section or key from DEFAULTS is not present in the config, it is
    added back, enforcing a minimal configuration.

    Returns
    -------
    ConfigParser
    """
    if _os.path.isfile(CONFIG_FILE):
        config.read(CONFIG_FILE)
    else:
        create_config()

    for section in DEFAULTS:
        if section not in _sections(config):
            config.add_section(section)
            _config_from_dict(config, DEFAULTS[section], section)
            write_config()
        for key, val in DEFAULTS[section].items():
            if key not in dict(config.items(section)):
                config.set(section, key, str(val))
                write_config()
    return config

def get_config():
    """Return a dictionary representation of the entire config."""
    return _config_to_dict(load_config())


def write_config():
    """Write the current config to CONFIG_FILE."""
    with open(CONFIG_FILE, 'w') as fout:
        config.write(fout)


###############################################################################
#                          Initialization Functions                           #
###############################################################################


def create_config_interactive(prompt=True):
    """Interact with the user to create a new config.

    Uses readline autocompletion to make setup easier.

    Parameters
    ----------
    prompt : bool
        As for confirmation before beginning wizard.
    """
    # Use tab completion
    t = _TabCompleter()
    _rl.set_completer_delims('\t')
    _rl.parse_and_bind("tab: complete")

    # Get permission
    if prompt:
        t.createListCompleter(['y', 'n'])
        _rl.set_completer(t.list_completer)
        print("Do you want to initialize your config at {}"
              .format(CONFIG_FILE))
        print("This will erase your current configuration (if it exists)")
        choice = _run.get_input("Initialize config? [y/N] ").strip().lower()
        if not choice == 'y':
            return

    cnf = DEFAULTS
    # Get path
    _rl.set_completer(_path_completer)
    # Not implemented yet
    #  print("\nThis module uses a database to store job information.",
          #  "This database should remain relatively small, but can get quite",
          #  "large if many jobs are submitted at once.\n"
          #  "It only needs to be accessible from the submit host, but should",
          #  "be somewhere with sufficient disk space (>500MB free).")
    #  print("Where would you like to put the db file?\n")
    #  file_path = _os.path.expanduser(
        #  _run.get_input('PATH: [{}] '.format(config.get('queue', 'db')))
    #  ).strip(' ').lower()

    #  file_path = file_path if file_path else cnf['queue']['db']
    #  cnf['queue']['db'] = _os.path.expanduser(file_path)

    print("We store job profile information in a small config file.")
    file_path = _os.path.expanduser(
        _run.get_input('Where would you like that file to go? [{}]'
                       .format(
                           _run.exp_file(config.get('jobs', 'profile_file'))
                       ))
    ).strip(' ').lower()

    file_path = file_path if file_path else cnf['jobs']['profile_file']
    cnf['jobs']['profile_file'] = _run.exp_file(file_path)


    # Temp file directory
    print("\nIt is possible to write all temporary and output files to a single",
          "temp file directory, regardless of where the job is run from.\n" +
          "As this library allows you to retrieve output files from the class",
          "directly, this is a good way of keeping your work directory tidy.",
          "If you do not choose to have files auto-deleted though, you will",
          "need to keep the directory tidy yourself.\n"
          "The value of this option is that if you run from a machine where",
          "some places are not cluster-accessible, jobs will run anyway if",
          "just one directory is accessible to the cluster.\n"
          "This option can always be overridden at runtime with the filepath",
          "argument.\nIf you leave the below option blank, the default will",
          "be to use the runtime path, which is usually the current working",
          "directory.\n")
    t.createListCompleter(['y', 'n'])
    _rl.set_completer(t.list_completer)
    if _run.get_yesno('Would you like to set a script file path?', 'y'):
        file_path = _get_set_path('Where would you like to put that file?')
        cnf['jobs']['scriptpath'] = file_path
    if _run.get_yesno('Would you like to set an output file path?', 'n'):
        file_path = _get_set_path('Where would you like to put that file?')
        cnf['jobs']['outpath'] = file_path

    # Cleaning
    t.createListCompleter(['y', 'n'])
    _rl.set_completer(t.list_completer)
    print('\nWe can automatically delete script and/or output files after',
          'results have been retrieved.\n'
          'This option can be overridden at run time on a per-job basis.\n'
          'Do you want to autoclean:\n')
    clean_files = _run.get_yesno('Autoclean script files?', 'y')
    cnf['jobs']['clean_files'] = clean_files

    clean_outs = _run.get_yesno('Autoclean output files (e.g. .out and .err)?',
                                'n')
    cnf['jobs']['clean_outputs'] = clean_outs

    # Wait times
    t.createListCompleter(cnf['queue'].values())
    _rl.set_completer(t.list_completer)
    max_len = _run.get_input("\nWhat is the maximum number of jobs allowed " +
                             "in your queue? [{}] "
                             .format(cnf['queue']['max_jobs']))
    max_len = max_len if max_len else cnf['queue']['max_jobs']
    cnf['queue']['max_jobs'] = int(max_len)

    # Options
    print('\nIs there a default queue you wish to submit to if no other',
          'options are given?\nIf so enter the name below, or leave blank',
          'to ignore.\n')
    def_queue = _run.get_input('Default queue: ').strip()

    print('\nThank you, configuring options.\n'
          'Please edit the file directly to inspect or edit your config.\n')

    create_config(cnf, def_queue)


def create_config(cnf=None, def_queue=None):
    """Create an initial config file.

    Gets all information from the file-wide DEFAULTS constant and overwrites
    specific keys using the values in cnf.

    This means that any records in the cnf dict that are not present in
    DEFAULTS will be ignored, and any records that are absent will be
    populated from DEFAULTS.

    Parameters
    ----------
    cnf : dict
        A dictionary of config defaults.
    def_queue : str
        A name for a queue to add to the default profile.
    """
    global config
    config = _configparser.ConfigParser(allow_no_value=True)

    if _os.path.exists(CONFIG_FILE):
        _os.remove(CONFIG_FILE)

    init_conf = {}

    if not cnf or not isinstance(cnf, dict):
        cnf = {}

    for section, items in DEFAULTS.items():
        init_conf[section] = {}
        if not section in cnf:
            cnf[section] = {}
        for key, value in items.items():
            val = cnf[section][key] if key in cnf[section] else value
            init_conf[section][key] = str(val)

    _config_from_dict(config, init_conf)

    with open(CONFIG_FILE, 'w') as fout:
        config.write(fout)

    if def_queue:
        def_prof = get_profile('DEFAULT')
        def_prof.args.update({'partition': def_queue})
        def_prof.write()


###############################################################################
#                           Config Helper Fuctions                            #
###############################################################################


def get_job_paths(kwds):
    """Parse keyword arguments to get important paths.

    Parameters
    ----------
    kwds : dict
        Keyword arguments accepted by fyrd.job.Job

    Returns
    -------
    kwds : dict
        A modified version of the input kwds dictionary with path requests
        removed
    runpath : sctr
        The path where the job will execute
    outpath : sctr
        The path where job outputs will go
    scriptpath : sctr
        The path where job script files will be stored
    """
    runpath = _os.path.abspath(kwds['dir'] if 'dir' in kwds else '.')
    kwds['dir'] = runpath

    # Set the output path
    cpath = get_option('jobs', 'outpath')
    if 'outpath' in kwds:
        outpath = kwds.pop('outpath')
    elif cpath:
        outpath = cpath
    else:
        outpath = runpath
    outpath = _os.path.abspath(outpath)

    # Set the script path
    cpath = get_option('jobs', 'scriptpath')
    if 'scriptpath' in kwds:
        scriptpath = kwds.pop('scriptpath')
    elif cpath:
        scriptpath = cpath
    else:
        scriptpath = outpath
    scriptpath = _os.path.abspath(scriptpath)

    return kwds, runpath, outpath, scriptpath


###############################################################################
#                                  Profiles                                   #
###############################################################################


class Profile(object):

    """A job submission profile. Just a thin wrapper around a dict.

    Attributes
    ----------
    name : str
    kwds : dict

    Methods
    -------
    write : Write self to config file
    """

    def __init__(self, name, kwds):
        """Set up bare minimum attributes.

        Parameters
        ----------
        name : str
            Name of the profile
        kwds : dict
            Dictionary of keyword arguments (will be validated).
        """
        name = 'DEFAULT' if name.lower() == 'default' else name
        self.name = name
        self.args = kwds

    def write(self):
        """Write self to config file."""
        set_profile(self.name, self.args)

    def __getattr__(self, key):
        """Access dict items as attributes."""
        if key in self.args:
            return self.args[key]

    def __setattr__(self, key, value):
        """Force checking of keywords."""
        if key == 'name':
            object.__setattr__(self, key, value)
        elif key == 'args':
            if not isinstance(value, dict):
                raise Exception('Keyword arguments must be a dict')
            value = _opt.check_arguments(value)
            object.__setattr__(self, key, value)
        else:
            opt, arg = list(_opt.check_arguments({key: value}).items())[0]
            self.args[opt] = arg

    def __len__(self):
        """Return arg count."""
        return len(self.args)

    def __repr__(self):
        """Display useful info."""
        return "{}<{}>".format(self.name, self.args)

    def __str__(self):
        """Pretty print."""
        title = 'DEFAULT' if self.name.lower() == 'default' else self.name
        return "{}:\n\t{}".format(
            title,
            '\n\t'.join(['{:<11}{:}'.format(i, j) \
                         for i, j in self.args.items()])
        )


def get_profile(profile=None, allow_none=True):
    """Return a profile if it exists, if None, return all profiles.

    Will return None if profile is supplied but does not exist.

    Parameters
    ----------
    profile : str
        The name of a profile to search for.
    allow_none : bool
        If True, return None if no profile matches, otherwise raise a
        ValueError.

    Returns
    -------
    fyrd.conf.Profile
        The requested profile.
    """
    load_profiles()
    # Allow lowercase default profile
    if profile and profile.lower() == 'default':
        profile = 'DEFAULT'
    if profile in _sections(profiles):
        return Profile(profile, _section_to_dict(profiles.items(profile)))
    else:
        if profile.lower() == 'default':
            _logme.log('default profile missing, recreating. You can '
                       'override the defaults by editing {}'
                       .format(CONFIG_FILE), 'warn')
            prof = Profile('default', DEFAULT_PROFILES['default'])
            prof.write()
            return prof
        if allow_none:
            return None
        else:
            raise ValueError('Requested profile ({}) does not exist'
                             .format(profile))


def get_profiles(profs=None, allow_none=True):
    """Return a dictionary of profiles from profiles.

    Returns all profiles if profiles argument is None.

    Parameters
    ----------
    profs : list
        A list of profiles to get.
    allow_none : bool
        If True, return None if no profile matches, otherwise raise a
        ValueError.

    Returns:
    dict:
        A ditionary of fyrd.conf.Profile objects
    """
    if profs:
        profs = _run.listify(profs)
        pfls = {}
        for profile in profs:
            pfls[profile] = get_profile(profile, allow_none)
    else:
        if not allow_none:
            raise ValueError('Profile required')
        pfls = {
            'DEFAULT': Profile(
                'DEFAULT', _section_to_dict(profiles.items('DEFAULT'))
            )
        }
        for section in _sections(profiles):
            pfls[section] = Profile(
                section, _section_to_dict(profiles.items(section))
            )
    return pfls


def set_profile(name, kwds, update=True):
    """Write profile to config file.

    Parameters
    ----------
    name : str
        The name of the profile to add/edit.
    kwds : dict
        Keyword arguments to add to the profile.
    update : bool
        Update the profile rather than overwriting it.
    """
    load_profiles()
    name = 'DEFAULT' if name.lower() == 'default' else name

    if not isinstance(kwds, dict):
        raise Exception('Profile arguments must be a dictionary')

    kwds = _opt.check_arguments(kwds)

    if name in _sections(profiles):
        if not update:
            _logme.log('Profile {} already exists, overwriting'.format(name),
                       'debug')
            profiles.remove_section(name)
            profiles.add_section(name)
    else:
        profiles.add_section(name)

    _config_from_dict(profiles, kwds, name)

    write_profiles()
    return get_profile(name)


def del_profile(name):
    """Delete a profile.

    Parameters
    ----------
    name : str
        The name of the profile to delete.
    """
    load_profiles()
    if name.lower() == 'default':
        for key in _section_to_dict(profiles.items('DEFAULT')):
            if key not in DEFAULT_PROFILES['DEFAULT']:
                profiles.remove_option('DEFAULT', key)
            else:
                profiles.set('DEFAULT', key,
                             str(DEFAULT_PROFILES['DEFAULT'][key]))
    elif name in _sections(profiles):
        _logme.log('Removing profile {}'.format(name))
        profiles.remove_section(name)
    else:
        _logme.log('Profile {} does not exist, cannot delete'.format(name),
                   'warn')
    write_profiles()


def create_profiles(profs=None):
    """Create an initial profiles file.

    Gets all information from the file-wide DEFAULT_PROFILES constant and
    overwrites specific keys using the values in cnf.

    This means that any records in the cnf dict that are not present in
    DEFAULT_PROFILES will be ignored, and any records that are absent will be
    populated from DEFAULT_PROFILES.

    Parameters
    ----------
    profs : dict
        A dictionary of config defaults.
    """
    global profiles

    profiles = _configparser.ConfigParser(
        defaults=DEFAULT_PROFILES['DEFAULT'],
        allow_no_value=True,
    )

    if _os.path.exists(_run.exp_file(config.get('jobs', 'profile_file'))):
        _os.remove(_run.exp_file(config.get('jobs', 'profile_file')))

    init_conf = {}
    if not profs or not isinstance(profs, dict):
        profs = {}
    for section, items in DEFAULT_PROFILES.items():
        init_conf[section] = {}
        if section not in profs:
            profs[section] = {}
        for key, value in items.items():
            val = profs[section][key] if key in profs[section] else value
            init_conf[section][key] = str(val)

    _config_from_dict(profiles, init_conf)

    with open(_run.exp_file(config.get('jobs', 'profile_file')), 'w') as fout:
        profiles.write(fout)


def load_profiles():
    """Load the profiles, create them if they don't exist.

    Returns
    -------
    ConfigParser
        profiles
    """
    if not _os.path.isfile(_run.exp_file(config.get('jobs', 'profile_file'))):
        create_profiles()
    profiles.read(_run.exp_file(config.get('jobs', 'profile_file')))

    # Recreate DEFAULT if necessary.
    def_prof = _config_to_dict(profiles)['DEFAULT']
    changed = False
    for key, val in DEFAULT_PROFILES['DEFAULT'].items():
        if key not in def_prof:
            profiles.set('DEFAULT', key, val)
            changed = True
    if changed:
        write_profiles()

    return profiles


def write_profiles():
    """Write the profiles to the file.

    Returns
    -------
    ConfigParser
        profiles
    """
    with open(_run.exp_file(config.get('jobs', 'profile_file')), 'w') as fout:
        profiles.write(fout)
    return load_profiles()


###############################################################################
#                              Helper Functions                               #
###############################################################################


def _sections(cnf, inc_anyway=False):
    """Include default in sections if it has items.

    Parameters
    ----------
    cnf : ConfigParser
        Any ConfigParser object
    inc_anyway : bool
        Include the DEFAULT section even if empty

    Returns
    -------
    list
        A list of sections in cnf, including DEFAULT if defined.
    """
    merged_sections = cnf.sections() + ['DEFAULT']
    return merged_sections if cnf.items('DEFAULT') or inc_anyway \
            else cnf.sections()


def _config_from_dict(cnf, dct, section=None):
    """Python 2 ConfigParsers cannot handle dictionaries, so we do it here.

    Parameters
    ----------
    cnf : ConfigParser
        A ConfigParser object.
    dct : dict
        A dictionary of {'key': 'value'} if section or {'section': {'key':
            'value'}}
    section : str
        An optional section name to update

    Returns
    -------
    ConfigParser
        The original ConfigParser object, updated.
    """
    assert isinstance(cnf, _configparser.ConfigParser)
    assert isinstance(dct, dict)

    if section:
        for v in dct.values():
            assert not isinstance(v, (tuple, list, dict, set))
    else:
        for v in dct.values():
            assert isinstance(v, dict)

    if section:
        if not section in cnf.sections() + ['DEFAULT']:
            cnf.add_section(section)
        for k, v in _dict_to_strings(dct).items():
            cnf.set(section, k, v)
    else:
        for sect, vals in dct.items():
            if not sect in cnf.sections() + ['DEFAULT']:
                cnf.add_section(sect)
            for k, v in _dict_to_strings(vals).items():
                cnf.set(sect, k, v)
    return cnf


def _dict_to_strings(kwds):
    """Convert all values in a dictionary to strings."""
    ot = {}
    assert isinstance(kwds, dict)
    for k, v in kwds.items():
        ot[k] = str(v)
    return ot


def _section_to_dict(section):
    """Convert a ConfigParser list of tuples to a dictionary.

    Parameters
    ----------
    section : list
        Output of ConfigParser.items(section)

    Returns
    -------
    dict
        A dictionary of the above with typed values
    """
    dct = dict(section)
    out = {}
    for key, val in dct.items():
        out[key] = _typecast_items(val)
    return out

def _typecast_items(x):
    """Try to convert a variable into the true type.

    Avoids eval() as we will be used on config files.

    For example: 'True' becomes True and '0.125' becomes 0.125
    """
    c = re.compile(r', *')
    r = re.compile(r': *')
    if not isinstance(x, (six.string_types, six.text_type)):
        return x
    if x == 'True':
        return True
    if x == 'False':
        return False
    if x == 'None':
        return None
    if x.isdigit():
        return int(x)
    try:
        return float(x)
    except (ValueError, TypeError):
        pass
    try:
        return complex(x)
    except (ValueError, TypeError):
        pass
    try:
        if x.startswith('[') and x.endswith(']'):
            return [_typecast_items(i) for i in c.split(x.strip('[]'))]
        if x.startswith('{') and x.endswith('}'):
            if ':' in x:
                return {
                    _typecast_items(k): _typecast_items(v) for k, v in [
                        r.split(i) for i in c.split(x.strip('{}'))
                    ]
                }
            return {_typecast_items(i) for i in c.split(x.strip('{}'))}
    except (ValueError, TypeError):
        pass
    if isinstance(x, (six.string_types)):
        return x.strip('\'"')
    return x


def _typecast_items(x):
    """Try to convert a variable into the true type.

    Avoids eval() as we will be used on config files.

    For example: 'True' becomes True and '0.125' becomes 0.125
    """
    c = re.compile(r', *')
    r = re.compile(r': *')
    if not isinstance(x, (six.string_types, six.text_type)):
        return x
    if x == 'True':
        return True
    if x == 'False':
        return False
    if x == 'None':
        return None
    if x.isdigit():
        return int(x)
    try:
        return float(x)
    except (ValueError, TypeError):
        pass
    try:
        return complex(x)
    except (ValueError, TypeError):
        pass
    try:
        if x.startswith('[') and x.endswith(']'):
            return [_typecast_items(i) for i in c.split(x.strip('[]'))]
        if x.startswith('{') and x.endswith('}'):
            if ':' in x:
                return {
                    _typecast_items(k): _typecast_items(v) for k, v in [
                        r.split(i) for i in c.split(x.strip('{}'))
                    ]
                }
            return {_typecast_items(i) for i in c.split(x.strip('{}'))}
    except (ValueError, TypeError):
        pass
    if isinstance(x, (six.string_types)):
        return x.strip('\'"')
    return x


def _config_to_dict(cnf):
    """Return a dictionary of all items from a ConfigParser object."""
    out = {}
    def_items = cnf.items('DEFAULT')
    if def_items:
        out['DEFAULT'] = _section_to_dict(def_items)
    for sect in _sections(cnf):
        out[sect] = _section_to_dict(cnf.items(sect))
    return out


###############################################################################
#                                 Completion                                  #
###############################################################################


class _TabCompleter(object):
    """
    A tab completer that can either complete from the filesystem or from a
    list.

    Taken from `https://gist.github.com/iamatypeofwalrus/5637895`_

    """
    list_completer = None


    def createListCompleter(self,ll):
        """
        This is a closure that creates a method that autocompletes from
        the given list.

        Since the autocomplete function can't be given a list to complete from
        a closure is used to create the list_completer function with a list to
        complete from.
        """

        def list_completer(_, state):
            """Make a completer from a list."""
            line   = _rl.get_line_buffer()

            if not line:
                return [c + " " for c in ll][state]

            else:
                return [c + " " for c in ll if c.startswith(line)][state]

        self.list_completer = list_completer


def _path_completer(_, state):
    """
    This is the tab completer for systems paths.
    Only tested on *nix systems
    """
    line = _rl.get_line_buffer()
    if not line:
        return _complete_path('.')[state]
    else:
        return _complete_path(line)[state]


def _listdir(root):
    "List directory 'root' appending the path separator to subdirs."
    res = []
    for name in _os.listdir(root):
        path = _os.path.join(root, name)
        if _os.path.isdir(path):
            name += _os.sep
        res.append(name)
    return res


def _complete_path(path=None):
    "Perform completion of filesystem path."
    if not path:
        return _os.listdir('.')
    path = _os.path.expanduser(path)
    dirname, rest = _os.path.split(path)
    tmp = dirname if dirname else '.'
    res = [_os.path.join(dirname, p)
           for p in _os.listdir(tmp) if p.startswith(rest)]
    # more than one match, or single match which does not exist (typo)
    if len(res) > 1 or not _os.path.exists(path):
        return res
    # resolved to a single directory, so return list of files below it
    if _os.path.isdir(path):
        return [_os.path.join(path, p) for p in _os.listdir(path)]
    # exact file match terminates this completion
    return [path + ' ']


def _get_set_path(message):
    """Get a path from the user, make the directory if necessary."""
    t = _TabCompleter()
    while True:
        _rl.set_completer(_path_completer)
        file_path = _run.get_input(message + ' ')
        file_path = file_path.strip().lower()
        if file_path:
            file_path = _os.path.abspath(file_path)
            if _os.path.isdir(file_path):
                break
            else:
                t.createListCompleter(['y', 'n'])
                _rl.set_completer(t.list_completer)
                ans = _run.get_input(
                    'That directory does not exist, would you like to ' +
                    'try to create it? [y/N] ')
                if ans.lower() == 'y':
                    try:
                        _os.makedirs(file_path)
                        break
                    except OSError:
                        print("Failed to make {}".format(file_path),
                              "Please make it manually or choose another",
                              "directory\n")
                        continue
                else:
                    continue
    return file_path


###############################################################################
#           Set Constants from config, initialize config on import            #
###############################################################################

# Create directory if it doesn't exist
if not _os.path.isdir(CONFIG_PATH):
    if _os.path.exists(CONFIG_PATH):
        _os.remove(CONFIG_PATH)
    _os.makedirs(CONFIG_PATH)
if not _os.path.isfile(CONFIG_FILE):
    create_config()

# Load config
config = load_config()

# Load profiles
profiles = load_profiles()



# -*- coding: utf-8 -*-
"""
Submit jobs to slurm or torque, or with multiprocessing.

 =============== ===================================================
         AUTHOR: Michael D Dacre, mike.dacre@gmail.com
   ORGANIZATION: Stanford University
        LICENSE: MIT License, property of Stanford, use as you wish
        VERSION: 0.6.2a1
        CREATED: 2015-12-11 22:19
  Last modified: 2017-08-17 14:47
 =============== ===================================================

Allows simple job submission with *dependency tracking and queue waiting* with
either torque or slurm. It uses simple techiques to avoid overwhelming the
queue and to catch bugs on the fly.

Setting Environment
-------------------

To set the environement, set batch_systems.MODE to one of the defined batch
systems, or run get_cluster_environment().

Simple Use
----------

At its simplest, this module can be used by just executing submit(<command>),
where command is a function or system command/shell script. The module will
autodetect the cluster, generate an intuitive name, run the job, and write all
outputs to files in the current directory. These can be cleaned with
clean_dir().

To run with dependency tracking, run::

    job  = submit(<command1>)
    job2 = submit(<command2>, dependencies=job1)

*NOTE:* To use this, the file that is calling submit must have all code inside
a function or protected by if __name__ == '__main__'.  In order to make
function submission work, the file that calls this script must be importable,
importing in python executes all code that isn't wrapped in an if __name__ ==
'__main__' clause. This can lead to infinite recursion!

Much more can be done though.

Class Overview
--------------

There are two important classes for interaction with the batch
system: Job and Queue. The essential flow of a job submission
is::

    job = Job(command/function, arguments, name)
    job.write()  # Writes the job submission files
    job.submit() # Submits the job
    job.wait()   # Waits for the job to complete
    job.stdout   # Prints the output from the job
    job.clean()  # Delete all of the files written

You can also wait for many jobs with the Queue class::

    q = Queue(user='self')
    q.wait([job1, job2])

The jobs in this case can be either a Job class or a job
number.

Profiles, Keywords, and the Config File
---------------------------------------

To make submission easier, this module defines a number of keyword arguments in
the options.py file that can be used for all submission and Job() functions.
These include things like 'cores' and 'nodes' and 'mem'. To avoid having to set
these every time, the module sets a config file at ~/.fyrd that
defines profiles. These can be edited directly in that file or through the
conf methods.

For example::

    conf.set_profile('small', {'nodes': 1, 'cores': 1, 'mem': '2GB'})

To see all profiles run::

    conf.get_profiles()

Other options are defined in the config file, including the maximum number of
jobs in the queue, the time to sleep between submissions, and other options. To
see these run::

    conf.get_option()

You can set options with::

    conf.set_option()

Feel free to alter the defaults in conf.py and options.py, they are
clearly documented.

Job Files
---------

All jobs write out a job file before submission, these files will end in
<suffix>.<job_suffix>, where suffix defaults to 'cluster' and can be set with
the 'suffix' keyword argument and <job_suffix> is defined in the batch_systems
config.

To change the directory these files are written to, use the 'filedir' keyword
argument to Job or submit.  *NOTE:* This *must* be accessible to the compute
nodes!!!

All jobs are assigned a name that is used to generate the output files,
including STDOUT and STDERR files. The default name for the out files is
STDOUT: name.cluster.out and STDERR: name.cluster.err. These can be overwridden
with keyword arguments.

Dependecy Tracking
------------------

Dependency tracking is supported in all modes.

To use dependency tracking in any mode pass a list of job ids to submit or
submit_file with the `dependencies` keyword argument.

Logging
-------

I use a custion logging script called logme to log errors. To get verbose
output, set logme.MIN_LEVEL to 'debug'. To reduce output, set logme.MIN_LEVEL
to 'warn'.

Help
----

Full help is available at: https://fyrd.rtfd.io
"""
import os as _os
import signal as _signal
import atexit as _atexit

#############
#  Version  #
#############

# Planned version
version = '0.6.2a1'
from ._version import get_versions
# Actual version (same as planned for releases)
__version__ = get_versions()['version']
del get_versions

###################
#  House Keeping  #
###################


class ClusterError(Exception):

    """A custom exception for cluster errors."""

    pass


#########################################
#  Make our functions easily available  #
#########################################


from . import queue
from . import job
from . import helpers
from . import batch_systems
from . import conf
from .run import check_pid as _check_pid

from .queue import Queue
from .batch_systems import check_queue
from .batch_systems import get_cluster_environment
from .batch_systems import options

from .job import Job
from .basic import submit
from .basic import submit_file
from .basic import make_job_file
from .basic import clean
from .basic import clean_dir
from .basic import wait
from .basic import get

from .helpers import jobify

from .conf import set_profile
from .conf import get_profile
from .conf import get_profiles

option_help = batch_systems.options.option_help

#  import fyrd.batch_system as batch_system

__all__ = ['Job', 'Queue', 'wait', 'get', 'submit', 'submit_file', 'jobify',
           'make_job_file', 'clean', 'clean_dir', 'check_queue', 'option_help',
           'set_profile', 'get_profile', 'get_profiles', 'conf', 'helpers']

##########################
#  Set the cluster type  #
##########################

batch_systems.MODE = get_cluster_environment()



# -*- coding: utf-8 -*-
"""
Define functions for using the Torque batch system
"""
import os as _os
import re as _re
from time import sleep as _sleep
import xml.etree.ElementTree as _ET
from subprocess import check_output as _check_output
from subprocess import CalledProcessError as _CalledProcessError

from .. import run as _run
from .. import conf as _conf
from .. import logme as _logme
from .. import ClusterError as _ClusterError
from .. import script_runners as _scrpts
from .. import submission_scripts as _sscrpt
_Script = _sscrpt.Script


PREFIX = '#PBS'
SUFFIX = 'qsub'

# Define torque-to-slurm mappings
TORQUE_SLURM_STATES = {
    'C': 'completed',
    'E': 'completing',
    'H': 'held',  # Not a SLURM state
    'Q': 'pending',
    'R': 'running',
    'T': 'suspended',
    'W': 'running',
    'S': 'suspended',
}


###############################################################################
#                             Functionality Test                              #
###############################################################################


def queue_test(warn=True):
    """Check that torque can be used.

    Just looks for qsub and qstat.

    Parameters
    ----------
    warn : bool
        log a warning on fail

    Returns
    -------
    batch_system_functional : bool
    """
    log_level = 'error' if warn else 'debug'
    qsub = _conf.get_option('queue', 'qsub')
    if qsub is not None and _os.path.dirname(qsub) and not _run.is_exe(qsub):
        _logme.log(
            'Cannot use torque as qsub path set in conf to {0}'
            .format(qsub) + ' but that path is not an executable',
            log_level
        )
        return False
    qsub = qsub if qsub else 'qsub'
    qsub = _run.which(qsub) if not _os.path.dirname(qsub) else qsub
    if not qsub:
        _logme.log(
            'Cannot use torque as cannot find qsub', log_level
        )
        return False
    qpath = _os.path.dirname(qsub)
    qstat = _os.path.join(qpath, 'qstat')
    return _run.is_exe(qstat)


###############################################################################
#                           Normalization Functions                           #
###############################################################################


def normalize_job_id(job_id):
    """Convert the job id into job_id, array_id."""
    job_id = job_id.split('.')[0]
    if '[' in job_id:
        job_id, array_id = job_id.split('[')
        job_id = job_id.strip('[]')
        array_id = array_id.strip('[]')
        if not array_id:
            array_id = None
    else:
        array_id = None
    return job_id, array_id


def normalize_state(state):
    """Convert state into standadized (slurm style) state."""
    if state.upper() in TORQUE_SLURM_STATES:
        state = TORQUE_SLURM_STATES[state.upper()]
    return state


###############################################################################
#                           Job Sumission Functions                           #
###############################################################################


def gen_scripts(job_object, command, args, precmd, modstr):
    """Create script object for job, does not create a sep. exec script.

    Parameters
    ---------
    job_object : fyrd.job.Job
    command : str
        Command to execute
    args : list
        List of additional arguments, not used in this script.
    precmd : str
        String from options_to_string() to add at the top of the file, should
        contain batch system directives
    modstr : str
        String to add after precmd, should contain module directives.

    Returns
    -------
    fyrd.script_runners.Script
        The submission script
    None
        Would be the exec_script, not used here.
    """
    scrpt = _os.path.join(
        job_object.scriptpath,
        '{0}.{1}.{2}'.format(job_object.name, job_object.suffix, SUFFIX)
    )

    sub_script = _scrpts.CMND_RUNNER_TRACK.format(
        precmd=precmd, usedir=job_object.runpath, name=job_object.name,
        command=command
    )
    return _Script(script=sub_script, file_name=scrpt), None


def submit(file_name, dependencies=None, job=None, args=None, kwds=None):
    """Submit any file with dependencies to Torque.

    Parameters
    ----------
    file_name : str
        Path to an existing torque submission file
    dependencies : list, optional
        List of torque job IDs as dependencies
    job : fyrd.job.Job, not implemented
        A job object for the calling job, not used by this functions
    args : list, not implemented
        A list of additional command line arguments to pass when submitting,
        not used by this function
    kwds : dict or str, not implemented
        A dictionary of keyword arguments to parse with options_to_string, or
        a string of option:value,option,option:value,.... Not used by this
        function.

    Returns:
        job_id (str)
    """
    _logme.log('Submitting to torque', 'debug')
    if dependencies:
        deps = '-W depend=afterok:{}'.format(
            ':'.join([str(d) for d in dependencies])
        )
        args = ['qsub', deps, file_name]
    else:
        args = ['qsub', file_name]

    # Try to submit job 5 times
    code, stdout, stderr = _run.cmd(args, tries=5)
    if code == 0:
        job_id, _ = normalize_job_id(stdout.split('.')[0])
    elif code == 17 and 'Unable to open script file' in stderr:
        _logme.log('qsub submission failed due to an already existing '
                   'script file, attempting to rename file and try '
                   'again.\nstderr: {}, stdout: {}, cmnd: {}'
                   .format(stderr, stdout, args), 'error')
        new_name = args[1] + '.resub'
        _os.rename(args[1], new_name)
        _logme.log('renamed script {} to {}, resubmitting'
                   .format(args[1], new_name), 'info')
        args[1] = new_name
        code, stdout, stderr = _run.cmd(args, tries=5)
        if code == 0:
            job_id, _ = normalize_job_id(stdout.split('.')[0])
        else:
            _logme.log('Resubmission still failed, aborting',
                       'critical')
            raise _CalledProcessError(code, args, stdout, stderr)
    else:
        if stderr.startswith('qsub: submit error ('):
            raise _ClusterError('qsub submission failed with error: ' +
                                '{}, command: {}'.format(stderr, args))
        else:
            _logme.log(
                'qsub failed with code {}\nstdout: {}\nstderr: {}'
                .format(code, stdout, stderr), 'critical'
            )
            raise _CalledProcessError(code, args, stdout, stderr)
    return job_id


###############################################################################
#                               Job Management                                #
###############################################################################


def kill(job_ids):
    """Terminate all jobs in job_ids.

    Parameters
    ----------
    job_ids : list or str
        A list of valid job ids or a single valid job id

    Returns
    -------
    success : bool
    """
    o = _run.cmd('qdel {0}'.format(' '.join(_run.listify(job_ids))), tries=5)
    return o[0] == 0


###############################################################################
#                           Queue Parsing Functions                           #
###############################################################################


def queue_parser(user=None, partition=None):
    """Iterator for torque queues.

    Use the `qstat -x -t` command to get an XML queue for compatibility.

    Parameters
    ----------
    user : str, optional
        User name to pass to qstat to filter queue with
    partiton : str, optional
        Partition to filter the queue with

    Yields
    ------
    job_id : str
    array_id : str or None
    name : str
    userid : str
    partition : str
    state :str
    nodelist : list
    numnodes : int
    cntpernode : int or None
    exit_code : int or Nonw

    cntpernode is currently always 1 as most torque queues treat every core as
    a node.
    """
    # I am not using run.cmd because I want to catch XML errors also
    try_count = 0
    qargs = ['qstat', '-x', '-t']
    r = _re.compile('<Variable_List>.*?</Variable_List>')
    while True:
        try:
            xmlstr = _check_output(qargs)
            try:
                xmlstr = xmlstr.decode()
            except AttributeError:
                pass
            # Get rid of the Variable_List as it is just the environment
            # and can sometimes have nonsensical characters.
            xmlstr = xmlstr.replace('\x1b', '')
            xmlstr = r.sub('', xmlstr)
            xmlqueue = _ET.fromstring(xmlstr)
        except _CalledProcessError:
            _sleep(1)
            if try_count == 5:
                raise
            else:
                try_count += 1
        except _ET.ParseError:
            # ElementTree throws error when string is empty
            _sleep(1)
            if try_count == 1:
                xmlqueue = None
                break
            else:
                try_count += 1
        else:
            break

    if xmlqueue is not None:
        for xmljob in xmlqueue:
            job_id, array_id = normalize_job_id(xmljob.find('Job_Id').text)
            job_owner = xmljob.find('Job_Owner').text.split('@')[0]
            if user and job_owner != user:
                continue
            job_name  = xmljob.find('Job_Name').text
            job_queue = xmljob.find('queue').text
            job_state = xmljob.find('job_state').text
            job_state = TORQUE_SLURM_STATES[job_state]
            _logme.log('Job {} state: {}'.format(job_id, job_state),
                       'debug')
            ndsx = xmljob.find('exec_host')
            if hasattr(ndsx, 'text') and ndsx.text:
                nds = ndsx.text.split('+')
            else:
                nds = []
            nodes = []
            # Convert node range to individual nodes in list
            for node in nds:
                if '/' in node:
                    nm, nums = node.split('/')
                    for num in nums.split(','):
                        if '-' in num:
                            nstart, nend = [int(i) for i in num.split('-')]
                        else:
                            nstart = nend = int(num)
                        nend += 1
                        for i in range(nstart, nend):
                            nodes.append(nm + '/' + str(i).zfill(2))
                else:
                    nodes.append(node)
            # I assume that every 'node' is a core, as that is the
            # default for torque, but it isn't always true
            job_threads  = len(nodes)
            exitcode     = xmljob.find('exit_status')
            if hasattr(exitcode, 'text'):
                exitcode = int(exitcode.text)
            else:
                exitcode = None

            if partition and job_queue != partition:
                continue

            # Torque doesn't have a variable scpu
            scpus = 1
            yield (job_id, array_id, job_name, job_owner, job_queue, job_state,
                   nodes, job_threads, scpus, exitcode)


def parse_strange_options(option_dict):
    """Parse all options that cannot be handled by the regular function.

    Parameters
    ----------
    option_dict : dict
        All keyword arguments passed by the user that are not already defined
        in the Job object

    Returns
    -------
    list
        A list of strings to be added at the top of the script file
    dict
        Altered version of option_dict with all options that can't be handled
        by `fyrd.batch_systems.options.option_to_string()` removed.
    None
        Would contain additional arguments to pass to qsub, but these are not
        needed so we just return None
    """
    outlist = []
    # Handle cores separately
    nodes = int(option_dict.pop('nodes')) if 'nodes' in option_dict else 1
    cores = int(option_dict.pop('cores')) if 'cores' in option_dict else 1

    outstring = '#PBS -l nodes={}:ppn={}'.format(nodes, cores)
    if 'features' in option_dict:
        outstring += ':' + ':'.join(
            _run.opt_split(option_dict.pop('features'), (',', ':')))
    if 'qos' in option_dict:
        outstring += ',qos={}'.format(option_dict.pop('qos'))
    outlist.append(outstring)

    return outlist, option_dict, None



#!/usr/bin/env python
"""
Submission strings for running remote commands.
"""

SCRP_RUNNER = """\
#!/bin/bash
{precmd}
mkdir -p $LOCAL_SCRATCH > /dev/null 2>/dev/null
if [ -f {script} ]; then
    {command}
    exit $?
else
    echo "{script} does not exist, make sure you set your filepath to a "
    echo "directory that is available to the compute nodes."
    exit 1
fi
"""

SCRP_RUNNER_TRACK = """\
#!/bin/bash
{precmd}
mkdir -p $LOCAL_SCRATCH > /dev/null 2>/dev/null
if [ -f {script} ]; then
    cd {usedir}
    date +'%y-%m-%d-%H:%M:%S'
    echo "Running {name}"
    {command}
    exitcode=$?
    echo Done
    echo Code: $exitcode
    date +'%y-%m-%d-%H:%M:%S'
    if [[ $exitcode != 0 ]]; then
        echo Exited with code: $exitcode >&2
    fi
    exit $exitcode
else
    echo "{script} does not exist, make sure you set your filepath to a "
    echo "directory that is available to the compute nodes."
    exit 1
fi
"""

CMND_RUNNER = """\
#!/bin/bash
{precmd}
mkdir -p $LOCAL_SCRATCH > /dev/null 2>/dev/null
cd {usedir}
{command}
exitcode=$?
if [[ $exitcode != 0 ]]; then
    echo Exited with code: $exitcode >&2
fi
exit $exitcode
"""

CMND_RUNNER_TRACK = """\
#!/bin/bash
{precmd}
mkdir -p $LOCAL_SCRATCH > /dev/null 2>/dev/null
cd {usedir}
date +'%y-%m-%d-%H:%M:%S'
echo "Running {name}"
{command}
exitcode=$?
echo Done
echo Code: $exitcode
date +'%y-%m-%d-%H:%M:%S'
if [[ $exitcode != 0 ]]; then
    echo Exited with code: $exitcode >&2
fi
exit $exitcode
"""

FUNC_RUNNER = r"""\
'''
Run a function remotely and pickle the result.

To try and make this as resistent to failure as possible, we import everything
we can, this sometimes results in duplicate imports and always results in
unnecessary imports, but given the context we don't care, as we just want the
thing to run successfully on the first try, no matter what.
'''
import os
import sys
import socket
from subprocess import Popen, PIPE
import six
from tblib import pickling_support
pickling_support.install()
import dill as pickle

out = None
try:
{imports}
{modimpstr}
except Exception:
    out = sys.exc_info()

ERR_MESSAGE = '''\
Failed to import your function. This usually happens when you have a module \
installed locally that is not available on the compute nodes.

In this case the module is {{}}.

However, I can only catch the first uninstalled module. To make sure all of \
your modules are installed on the compute nodes, do this::

    freeze --local | grep -v ^\-e | cut -d = -f 1 "> module_list.txt\n

Then, submit a job to the compute nodes with this command::

    cat module_list.txt | xargs pip install --user
'''


def run_function(func_c, args=None, kwargs=None):
    '''Run a function with arglist and return output.'''
    if not hasattr(func_c, '__call__'):
        raise Exception('{{}} is not a callable function.'
                        .format(func_c))
    if args and kwargs:
        ot = func_c(*args, **kwargs)
    elif args:
        try:
            iter(args)
        except TypeError:
            args = (args,)
        if isinstance(args, (six.string_types, six.text_type)):
            args = (args,)
        ot = func_c(*args)
    elif kwargs:
        ot = func_c(**kwargs)
    else:
        ot = func_c()
    return ot


if __name__ == "__main__":
    # If an Exception was raised during import, skip this
    if not out:
        with open('{pickle_file}', 'rb') as fin:
            # Try to install packages first
            try:
                function_call, args, kwargs = pickle.load(fin)
            except ImportError as e:
                out = sys.exc_info()
                module = str(e).split(' ')[-1]
                node   = socket.gethostname()
                sys.stderr.write(ERR_MESSAGE.format(module))
                out = list(out)
                out[1] = ImportError(
                    'Module {{}} is not installed on compute node {{}}'
                    .format(module, node)
                )
                out = tuple(out)
            except:
                out = sys.exc_info()

    try:
        if not out:
            out = run_function(function_call, args, kwargs)
    except Exception:
        out = sys.exc_info()

    with open('{out_file}', 'wb') as fout:
        pickle.dump(out, fout)

    if isinstance(out, tuple):
        if issubclass(BaseException, out[0]):
            six.reraise(*out)
"""



# -*- coding: utf-8 -*-
"""
Classes to build submission scripts.
"""
import os  as _os
import sys as _sys
import inspect as _inspect
import dill as _pickle

###############################################################################
#                               Import Ourself                                #
###############################################################################

from . import run as _run
from . import logme as _logme
from . import script_runners as _scrpts


class Script(object):

    """A script string plus a file name."""

    written = False

    def __init__(self, file_name, script):
        """Initialize the script and file name."""
        self.script    = script
        self.file_name = _os.path.abspath(file_name)

    def write(self, overwrite=True):
        """Write the script file."""
        _logme.log('Script: Writing {}'.format(self.file_name), 'debug')
        pth = _os.path.split(_os.path.abspath(self.file_name))[0]
        if not _os.path.isdir(pth):
            raise OSError('{} Does not exist, cannot write scripts'
                          .format(pth))
        if overwrite or not _os.path.exists(self.file_name):
            with open(self.file_name, 'w') as fout:
                fout.write(self.script + '\n')
            self.written = True
            return self.file_name
        else:
            return None

    def clean(self, delete_output=None):
        """Delete any files made by us."""
        if delete_output:
            _logme.log('delete_output not implemented in Script', 'debug')
        if self.written and self.exists:
            _logme.log('Script: Deleting {}'.format(self.file_name), 'debug')
            _os.remove(self.file_name)

    @property
    def exists(self):
        """True if file is on disk, False if not."""
        return _os.path.exists(self.file_name)

    def __repr__(self):
        """Display simple info."""
        return "Script<{}(exists: {}; written: {})>".format(
            self.file_name, self.exists, self.written)

    def __str__(self):
        """Print the script."""
        return repr(self) + '::\n\n' + self.script + '\n'


class Function(Script):

    """A special Script used to run a function."""

    def __init__(self, file_name, function, args=None, kwargs=None,
                 imports=None, syspaths=None, pickle_file=None, outfile=None):
        """Create a function wrapper.

        NOTE: Function submission will fail if the parent file's code is not
        wrapped in an if __main__ wrapper.

        Parameters
        ----------
        file_name : str
            A root name to the outfiles
        function : callable
            Function handle.
        args : tuple, optional
            Arguments to the function as a tuple.
        kwargs : dict, optional
            Named keyword arguments to pass in the function call
        imports : list, optional
            A list of imports, if not provided, defaults to all current
            imports, which may not work if you use complex imports.  The list
            can include the import call, or just be a name, e.g ['from os
            import path', 'sys']
        syspaths : list, optional
            Paths to be included in submitted function
        pickle_file : str, optional
            The file to hold the function.
        outfile : str, optional
            The file to hold the output.
        """
        _logme.log('Building Function for {}'.format(function), 'debug')
        self.function = function
        self.parent   = _inspect.getmodule(function)
        self.args     = args
        self.kwargs   = kwargs

        ##########################
        #  Take care of imports  #
        ##########################
        filtered_imports = _run.get_all_imports(
            function, {'imports': imports}, prot=True
        )

        # Get rid of duplicates and join imports
        impts = _run.indent('\n'.join(set(filtered_imports)), '    ')

        # Import the function itself
        func_import = _run.indent(_run.import_function(function), '    ')

        # sys paths
        if syspaths:
            _logme.log('Syspaths: {}'.format(syspaths), 'debug')
            impts = (_run.indent(_run.syspath_fmt(syspaths), '    ') + '\n\n'
                     + impts)

        # Set file names
        self.pickle_file = pickle_file if pickle_file else file_name + '.pickle.in'
        self.outfile     = outfile if outfile else file_name + '.pickle.out'

        # Create script text
        script = '#!{}\n'.format(_sys.executable)
        script += _scrpts.FUNC_RUNNER.format(name=file_name,
                                             modimpstr=func_import,
                                             imports=impts,
                                             pickle_file=self.pickle_file,
                                             out_file=self.outfile)

        super(Function, self).__init__(file_name, script)

    def write(self, overwrite=True):
        """Write the pickle file and call the parent Script write function."""
        _logme.log('Writing pickle file {}'.format(self.pickle_file), 'debug')
        with open(self.pickle_file, 'wb') as fout:
            _pickle.dump((self.function, self.args, self.kwargs), fout)
        super(Function, self).write(overwrite)

    def clean(self, delete_output=False):
        """Delete the input pickle file and any scripts.

        Parameters
        ----------
        delete_output : bool, optional
            Delete the output pickle file too.
        """
        if self.written:
            if _os.path.isfile(self.pickle_file):
                _logme.log('Function: Deleting {}'.format(self.pickle_file),
                           'debug')
                _os.remove(self.pickle_file)
            else:
                _logme.log('Function: {} already gone'
                           .format(self.pickle_file), 'debug')
            if delete_output:
                if _os.path.isfile(self.outfile):
                    _logme.log('Function: Deleting {}'.format(self.outfile),
                               'debug')
                    _os.remove(self.outfile)
                else:
                    _logme.log('Function: {} already gone'
                               .format(self.outfile), 'debug')
        super(Function, self).clean(None)



# -*- coding: utf-8 -*-
"""
User notification using email, either Linux or SMTP.

Must be configured in the primary config file.

Primary function
----------------
notify(msg, to, subject)
"""
import os as _os
import smtplib as _smtplib
import subprocess as _sub

from six.moves import email_mime_multipart as _multi
from six.moves import email_mime_text as _mtxt

from fyrd import run as _run
from fyrd import conf as _conf
from fyrd import logme as _logme


def notify(msg, to=None, subject='Fyrd notification'):
    """Send a message."""
    if not to:
        to = _conf.get_option('notify', 'notify_address')
    if not to:
        _logme.log('No notification address, cannot send notification',
                   'error')
        return False
    assert isinstance(msg, str)
    assert isinstance(to, str)
    assert isinstance(subject, str)
    opt = _conf.get_option('notify', 'mode', None)
    if opt == 'linux':
        return mail(msg, to, subject)
    elif opt == 'smtp':
        return smtp(msg, to, subject)
    raise ValueError('Unknown mail mode {}'.format(opt))


def mail(msg, to, subject):
    """Send a message with linux mail."""
    _logme.log('Sending with sendmail', 'debug')
    if not _run.which('mail'):
        _logme.log('Linux mail not found, cannot send notification' +
                   'perhaps configure SMTP?', 'error')
        return False
    job = _sub.run(
        ['mail', '-s', subject, to],
        input=msg.encode(),
        stdout=_sub.PIPE, stderr=_sub.PIPE
    )
    if not job.returncode == 0:
        raise _sub.CalledProcessError(
            cmd='mail -s {} {}'.format(subject, to),
            output=job.stdout, stderr=job.stderr,
            returncode=job.returncode
        )


def smtp(msg, to, subject):
    """Send a message via smtp."""
    _logme.log('Sending with smtp', 'debug')
    # Get Options
    host = _conf.get_option('notify', 'smtp_host', None)
    if not host:
        raise ValueError('smtp_host must be specified in the config')
    port = int(_conf.get_option('notify', 'smtp_port', 587))
    tls = _conf.get_option('notify', 'smtp_tls', True)
    fromaddr = _conf.get_option('notify', 'smtp_from', None)
    if not fromaddr:
        raise ValueError('smtp_from must be specified in the config')
    user = _conf.get_option('notify', 'smtp_user', None)
    if not user:
        user = fromaddr
    passfl = _conf.get_option('notify', 'smtp_passfile', None)
    if not passfl:
        raise ValueError('smtp_passfl must be specified in the config')
    with open(_os.path.expanduser(passfl)) as fin:
        passwd = fin.readline().strip()

    # Prepare Message
    body = msg
    msg = _multi.MIMEMultipart()
    msg['From'] = fromaddr
    msg['To'] = to
    msg['Subject'] = subject

    msg.attach(_mtxt.MIMEText(body, 'plain'))

    server = _smtplib.SMTP(host, port)
    if tls:
        server.starttls()
    server.login(user, passwd)
    text = msg.as_string()
    server.sendmail(fromaddr, to, text)
    server.quit()



# -*- coding: utf-8 -*-
"""
Setup Script for Fyrd
"""
import os
import codecs

import setuptools
from setuptools import setup
from setuptools.command.test import test as TestCommand
import versioneer

log = setuptools.distutils.log

VERSION=versioneer.get_version()
GITHUB='https://github.com/MikeDacre/fyrd'

###############################################################################
#                            A class to run tests                             #
###############################################################################

class TestRunner(TestCommand):

    """Run script in tests directory with py.test and without.

    The local queue can't be tested with py.test, so we run it outside of
    py.test, we also skip remote tests here because they require some config
    before being able to successfully execute.
    """

    def run_tests(self):
        """Run the test script, skip remote tests here."""
        import sys
        from subprocess import check_call
        # The remote queue testing can fail for a variety of config reasons
        # so we won't run it here
        check_call([sys.executable, 'tests/run_tests.py', '-l'])


###############################################################################
#                     Build the things we need for setup                      #
###############################################################################

# Get the long description from the README file
here = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Generate a list of python scripts
scpts = []
scpt_dir = os.listdir(os.path.join(here, 'bin'))
for scpt in scpt_dir:
    scpts.append(os.path.join('bin', scpt))

# Set command class
cmdclass = versioneer.get_cmdclass()
tcmdclss = {'test': TestRunner}
for k, v in cmdclass.items():
    tcmdclss[k] = v
cmdclass = tcmdclss

###############################################################################
#                                Setup Options                                #
###############################################################################

setup(
    name='fyrd',
    version=VERSION,
    description=('Submit functions and shell scripts to torque, slurm, ' +
                 'or local machines'),
    long_description=long_description,
    url='https://fyrd.science',
    download_url='{}/archive/v{}.tar.gz'.format(
        GITHUB, VERSION
    ),
    author='Michael Dacre',
    author_email='mike.dacre@gmail.com',
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Environment :: Console',
        'Operating System :: POSIX :: Linux',
        'Natural Language :: English',
        'Topic :: System :: Clustering',
        'Topic :: System :: Monitoring',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='slurm torque multiprocessing cluster job_management',

    requires=['dill', 'tabulate', 'six', 'tblib', 'psutil'
              'tqdm', 'Pyro4', 'sqlalchemy'],
    install_requires=['dill', 'tabulate', 'six', 'Pyro4', 'psutil',
                      'tblib', 'tqdm', 'sqlalchemy'],
    tests_require=['pytest'],
    packages=['fyrd', 'fyrd/batch_systems'],
    cmdclass=cmdclass,
    scripts=scpts,
    entry_points={
        'console_scripts': [
            'fyrd = fyrd.__main__:main',
        ]
    },
)



# -*- coding: utf-8 -*-
"""
Available options for job submission.

All keyword arguments that can be used with Job() objects are defined in this
file. These can be edited by the end user to increase functionality.

Options are defined in dictionaries with the syntax::

    {'name': {
        'slurm': The command to be used for slurm
        'torque': The command to be used for torque
        'default': The default to use if not set
        'type': The python object type for the option
        'help': A string with help information
        }
    }

All of these fields are required except in the case that:

1. The option is managed in options_to_string explicitly
2. The option is in TORQUE or SLURM dictionaries, in which case flags used by
   other queue systems can be skipped.
"""
import os  as _os
import re  as _re
import sys as _sys
from textwrap import wrap as _wrap
from itertools import groupby as _groupby
from collections import OrderedDict as _OD

from six import reraise as _raise
from six import text_type as _txt
from six import string_types as _str
from tabulate import tabulate as _tabulate

from .. import run
from .. import logme
from .. import ClusterError

from . import MODE
from . import get_batch_system
from . import check_queue

__all__ = ['option_help']

###############################################################################
#                       Possible Job Submission Options                       #
###############################################################################

# Defined by dictionaries:
#      type: python type to convert option into
#     slurm: the string to format args into for slurm
#    torque: the string to format args into for torque
# [s|t]join: used to join list types for slurm or torque
#      help: Info for the user on the option

# Options available in all modes
COMMON  = _OD([
    ('depends',
     {'help': 'A job or list of jobs to depend on',
      'default': None, 'type': list}),
    ('clean_files',
     {'help': 'Auto clean script files when fetching outputs',
      'default': None, 'type': bool}),
    ('clean_outputs',
     {'help': 'Auto clean output files when fetching outputs',
      'default': None, 'type': bool}),
    ('cores',
     {'help': 'Number of cores to use for the job',
      'default': 1, 'type': int}),
    ('modules',
     {'help': 'Modules to load with the `module load` command',
      'default': None, 'type': list}),
    ('imports',
     {'help': 'Imports to be used in function calls (e.g. sys, os)',
      'default': None, 'type': list}),
    ('syspaths',
     {'help': 'Paths to add to _sys.path for submitted functions',
      'default': None, 'type': list}),
    ('scriptpath',
     {'help': 'Folder to write cluster script files to, must be accessible ' +
              'to the compute nodes.',
      'default': '.', 'type': str}),
    ('outpath',
     {'help': 'Folder to write cluster output files to, must be accessible ' +
              'to the compute nodes.',
      'default': '.', 'type': str}),
    ('runpath',
     {'help': 'The working directory for the job',
      'default': '.', 'type': str,
      'slurm': '--workdir={}', 'torque': '-d {}'}),
    ('suffix',
     {'help': 'A suffix to append to job files (e.g. job.suffix.qsub)',
      'default': 'cluster', 'type': str}),
    ('outfile',
     {'help': 'File to write STDOUT to',
      'default': None, 'type': str,
      'slurm': '-o {}', 'torque': '-o {}'}),
    ('errfile',
     {'help': 'File to write STDERR to',
      'default': None, 'type': str,
      'slurm': '-e {}', 'torque': '-e {}'}),
])

# Options used in all batch systems
CLUSTER_CORE = _OD([
    ('nodes',
     {'help': 'Number of nodes to request',
      'default': 1, 'type': int}),
    ('features',
     {'help': 'A comma-separated list of node features to require',
      'slurm': '--constraint={}',  # Torque in options_to_string()
      'default': None, 'type': list, 'sjoin': '&'}),
    ('qos',
     {'help': 'A quality of service to require',
      'slurm': '--qos={}',  # Torque in options_to_string()
      'default': None, 'type': str}),
    ('time',
     {'help': 'Walltime in HH:MM:SS',
      'default': '12:00:00', 'type': str,
      'slurm': '--time={}', 'torque': '-l walltime={}'}),
    # We explictly set MB in torque
    ('mem',
     {'help': 'Memory to use in MB (e.g. 4000)',
      'default': 4000, 'type': (int, str),
      'slurm': '--mem={}', 'torque': '-l mem={}MB'}),
    ('partition',
     {'help': 'The partition/queue to run in (e.g. local/batch)',
      'default': None, 'type': str,
      'slurm': '-p {}', 'torque': '-q {}'}),
])

# Note: There are many more options, as them as need to the following lists,
#       CLUSTER_OPTS should be used for options that work on both systems,
#       the TORQUE and SLURM dictionaries should be used for options that are
#       unique to one.

# Additional options shared between systems
CLUSTER_OPTS = _OD([
    ('account',
     {'help': 'Account to be charged', 'default': None, 'type': str,
      'slurm': '--account={}', 'torque': '-A {}'}),
    ('export',
     {'help': 'Comma separated list of environmental variables to export',
      'default': None, 'type': str,
      'slurm': '--export={}', 'torque': '-v {}'}),
])

###############################################################################
#                                Torque Options                               #
#  from: adaptivecomputing.com/torque/4-0-2/Content/topics/commands/qsub.htm  #
###############################################################################

TORQUE = _OD()

#####################################################
#                   SLURM Options                   #
#  from: http://slurm.schedmd.com/pdfs/summary.pdf  #
#####################################################

SLURM  = _OD([
    ('begin',
     {'help': 'Start after this much time',
      'slurm': '--begin={}', 'type': str,
      'default': None}),
])

################################################################
#                         SYNONYMS                             #
#  These allow alternate keyword arguments for common options  #
################################################################


SYNONYMS = _OD([
    ('depend',         'depends'),
    ('dependency',     'depends'),
    ('dependencies',   'depends'),
    ('stdout',         'outfile'),
    ('stderr',         'errfile'),
    ('queue',          'partition'),
    ('memory',         'mem'),
    ('cpus',           'cores'),
    ('threads',        'cores'),
    ('walltime',       'time'),
    ('delete_files',   'clean_files'),
    ('delete_outputs', 'clean_outputs'),
    ('filedir',        'scriptpath'),
    ('filepath',       'scriptpath'),
    ('dir',            'runpath'),
    ('path',           'runpath'),
    ('paths',          'syspaths'),
    ('syspath',        'syspaths'),
    ('scriptdir',      'scriptpath'),
    ('cleanfiles',     'clean_files'),
    ('delfiles',       'clean_files'),
    ('cleanouts',      'clean_outputs'),
    ('delouts',        'clean_outputs'),
    ('deloutputs',     'clean_outputs'),
    ('cleanoutputs',   'clean_outputs'),
])


###############################################################################
#                       DO NOT EDIT BELOW THIS LINE!!!                        #
###############################################################################


###############################################################################
#                     Composites for Checking and Lookup                      #
###############################################################################


SLURM_KWDS = COMMON.copy()
for kds in [CLUSTER_CORE, CLUSTER_OPTS, SLURM]:
    SLURM_KWDS.update(kds)

TORQUE_KWDS = COMMON.copy()
for kds in [CLUSTER_CORE, CLUSTER_OPTS, TORQUE]:
    TORQUE_KWDS.update(kds)

CLUSTER_KWDS = SLURM_KWDS.copy()
CLUSTER_KWDS.update(TORQUE_KWDS)

# Should include the above in a dictionary by qtype
BATCH_KWDS = {
    'slurm': SLURM_KWDS,
    'torque': TORQUE_KWDS,
}

ALL_KWDS = CLUSTER_KWDS.copy()

# Will be 'name' -> type
ALLOWED_KWDS = _OD()
for name, info in ALL_KWDS.items():
    ALLOWED_KWDS[name] = info['type'] if 'type' in info else None


###############################################################################
#                      Option Handling Custom Exception                       #
###############################################################################

class OptionsError(ClusterError):

    """A custom Exception for failures in option parsing."""

    pass


###############################################################################
#                          Option Handling Functions                          #
###############################################################################


def sanitize_arguments(kwds):
    """Run check_arguments, but return unmatched keywords as is."""
    new_kwds = dict()
    for opt, arg in kwds.items():
        try:
            o, a = list(check_arguments({opt: arg}).items())[0]
            new_kwds[o] = a
        except OptionsError:
            new_kwds[opt] = arg
    return new_kwds


def split_keywords(kwargs):
    """Split a dictionary of keyword arguments into two dictionaries.

    The first dictionary will contain valid arguments for fyrd, the second will
    contain all others.

    Returns
    -------
    valid_args, other_args: dict
    """
    if not isinstance(kwargs, dict):
        raise ValueError('Invalid argument. Should be a dictionary, is {}'
                         .format(type(kwargs)))
    good = {}
    bad  = {}
    for key, val in kwargs.items():
        try:
            good.update(check_arguments({key: val}))
        except OptionsError:
            bad.update({key: val})
    return check_arguments(good), bad


def check_arguments(kwargs):
    """Make sure all keywords are allowed.

    Raises OptionsError on error, returns sanitized dictionary on success.

    Note: Checks in SYNONYMS if argument is not recognized, raises OptionsError
          if it is not found there either.
    """
    new_kwds = {}
    # Make sure types are correct
    for arg, opt in kwargs.items():
        if arg not in ALLOWED_KWDS:
            if arg in SYNONYMS:
                arg = SYNONYMS[arg]
                assert arg in ALLOWED_KWDS
            else:
                raise OptionsError('Unrecognized argument {}'.format(arg))
        if opt is not None and not isinstance(opt, ALLOWED_KWDS[arg]):
            newtype = ALLOWED_KWDS[arg]
            if (newtype is list or newtype is tuple) \
                    and not isinstance(arg, (list, tuple)):
                opt = run.listify(opt)
            elif newtype is int and isinstance(opt, str) and opt.isdigit():
                opt = int(opt)
            else:
                raise TypeError(
                    'arg "{}" must be {}, is {} ({})'.format(
                        arg, ALLOWED_KWDS[arg], opt, type(opt)
                    )
                )
        new_kwds[arg] = opt

    # Parse individual complex options
    for arg, opt in new_kwds.items():
        if arg == 'time':
            try:
                if '-' in opt:
                    day, time = opt.split('-')
                else:
                    day = 0
                    time = opt
                time = [int(i) for i in time.split(':')]
                if len(time) == 3:
                    hours, mins, secs = time
                elif len(time) == 2:
                    hours = 0
                    mins, secs = time
                elif len(time) == 1:
                    hours = mins = 0
                    secs = time[0]
                hours = (int(day)*24) + hours
                if secs > 60:
                    mins += 1
                    secs = secs%60
                if mins > 60:
                    hours += 1
                    mins = mins%60
                opt = '{}:{}:{}'.format(str(hours).rjust(2, '0'),
                                        str(mins).rjust(2, '0'),
                                        str(secs).rjust(2, '0'))
                new_kwds[arg] = opt
            except:
                raise OptionsError('time must be formatted as D-HH:MM:SS ' +
                                   'or a fragment of that (e.g. MM:SS) ' +
                                   'it is formatted as {}'.format(opt))

        # Force memory into an integer of megabytes
        elif arg == 'mem' and isinstance(opt, (_str, _txt)):
            if opt.isdigit():
                opt = int(opt)
            else:
                # Try to guess unit by suffix
                memerror = ('mem is malformatted, should be a number '
                            'of MB or a string like 24MB or 10GB, '
                            'it is: {}'.format(opt))
                groups = _groupby(opt, key=str.isdigit)
                try:
                    svalk, svalg = next(groups)
                    sval  = int(''.join(svalg))
                    sunitk, sunitg = next(groups)
                    sunit = ''.join(sunitg).lower()
                except ValueError:
                    err = list(_sys.exc_info())
                    err[1] = ValueError(memerror)
                    _raise(*err)
                if list(groups) or not svalk or sunitk:
                    raise ValueError(memerror)
                if sunit == 'b':
                    opt = int(float(sval)/float(1024)/float(1024))
                elif sunit == 'kb' or sunit == 'k':
                    opt = int(float(sval)/float(1024))
                elif sunit == 'mb' or sunit == 'm':
                    opt = sval
                elif sunit == 'gb' or sunit == 'g':
                    opt = sval*1024
                elif sunit == 'tb' or sunit == 't':
                    # Crazy people
                    opt = sval*1024*1024
                else:
                    raise ValueError('Unknown memory unit opt {}'
                                     .format(sunit))
                # Don't allow 0, minimum memory req is 5MB
                if opt < 5:
                    opt = 5
            new_kwds[arg] = opt

    return new_kwds


def option_to_string(option, value=None, qtype=None):
    """Return a string with an appropriate flag for slurm or torque.

    Parameters
    ----------
    option : str
        An allowed option definied in options.all_options
    value : str, optional
        A value for that option if required (if None, default used)
    qtype : str, optional
        One of the defined batch systems

    Returns
    -------
    str
        A string with the appropriate flags for the active queue.
    """
    # Import a couple of queue functions here
    qtype = qtype if qtype else MODE

    if isinstance(option, dict):
        raise ValueError('Arguments to option_to_string cannot be '
                         'dictionaries, you probably want options_to_string')

    option = str(option).rstrip()

    if option == 'cores' or option == 'nodes':
        raise OptionsError('Cannot handle cores or nodes here, use ' +
                           'options_to_string')

    kwds = BATCH_KWDS[qtype]

    # Make sure argument allowed
    option, value = list(check_arguments({option: value}).items())[0]

    # Fail with debug error if option not available in this mode
    if option in ALLOWED_KWDS and option not in kwds:
        logme.log('{} not available in {} mode.'.format(option, qtype),
                  'debug')
        return ''

    # Try to get default
    if not value:
        if not kwds[option]['type'] == bool:
            if 'default' in kwds[option]:
                value = kwds[option]['default']
                logme.log('Using default value {} for {}'
                          .format(value, option), 'debug')
            else:
                raise OptionsError('{} requires a value'.format(option))

    # Return formatted string
    prefix = get_batch_system(qtype).PREFIX
    if '{}' in kwds[option][qtype]:
        if value is None:
            return ''
        if 'type' in kwds[option]:
            if not isinstance(value, kwds[option]['type']):
                raise ValueError(
                    'Argument to {} must be {}, but is:\n{}'
                    .format(option, kwds[option]['type'], value)
                )
        return '{prefix} {optarg}'.format(
            prefix=prefix, optarg=kwds[option][qtype].format(value))
    else:
        return '{prefix} {option}'.format(prefix=prefix,
                                          option=kwds[option][qtype])


def options_to_string(option_dict, qtype=None):
    """Return a multi-line string for job submission.

    This function pre-parses options and then passes them to the
    parse_strange_options function of each batch system, before using the
    option_to_string function to parse the remaining options.

    Parameters
    ----------
    option_dict : dict
        Dict in format {option: value} where value can be None. If value is
        None, default used.
    qtype : str
        The defined batch system

    Returns
    -------
    parsed_options : str
        A multi-line string of parsed options
    runtime_options : list
        A list of parsed options to be used at submit time
    """
    qtype = qtype if qtype else MODE
    batch = get_batch_system(qtype)
    check_queue(qtype)

    # Sanitize arguments
    if not isinstance(option_dict, dict):
        raise TypeError('option_dict must be dict is {}'.format(
            type(option_dict)))

    option_dict = check_arguments(option_dict.copy())

    outlist = []

    # Set path if required
    if 'filepath' in option_dict:
        filepath = _os.path.abspath(option_dict.pop('filepath'))
        if 'outfile' in option_dict:
            option_dict['outfile'] = _os.path.join(
                filepath, _os.path.basename(option_dict['outfile']))
        if 'errfile' in option_dict:
            option_dict['errfile'] = _os.path.join(
                filepath, _os.path.basename(option_dict['errfile']))

    outlist, option_dict, other_args = batch.parse_strange_options(option_dict)

    # Loop through all remaining options
    for option, value in option_dict.items():
        outlist.append(option_to_string(option, value, qtype))

    optstring = _re.sub(r'[\n]+', '\n', '\n'.join(outlist))

    return optstring, other_args


def option_help(mode='string', qtype=None, tablefmt='simple'):
    """Print a sting to stdout displaying information on all options.

    The possible run modes for this extension are:

    ============ =========================================
    string       Return a formatted string
    print        Print the string to stdout
    list         Return a simple list of keywords
    table        Return a table of lists
    merged_table Combine all keywords into a single table
    ============ =========================================

    Parameters
    ----------
    mode : {'string', 'print', 'list', 'table', 'merged_table'}, optional
    qtype : str, optional
        If provided only return info on that queue type.
    tablefmt : str, optional
        A tabulate-style table format, one of::

            'plain', 'simple', 'grid', 'pipe', 'orgtbl',
            'rst', 'mediawiki', 'latex', 'latex_booktabs'

    Returns
    -------
    str
        A formatted string
    """

    hlp = _OD()

    # Explicitly get the function call help out of core to treat separately
    common = COMMON.copy()
    impts  = common.pop('imports')

    hlp['common'] = {
        'summary': 'Used in every mode',
        'help': common,
    }

    hlp['func'] = {
        'summary': 'Used for function calls',
        'help': _OD([('imports', impts)]),
    }

    # Include all cluster options in one
    cluster = CLUSTER_CORE.copy()
    cluster.update(CLUSTER_OPTS)
    hlp['cluster'] = {
        'summary': 'Options that work in both slurm and torque',
        'help': cluster,
    }

    if TORQUE:
        hlp['torque'] = {
            'summary': "Used for torque only",
            'help': TORQUE,
        }

    if SLURM:
        hlp['slurm'] = {
            'summary': "Used for slurm only",
            'help': SLURM,
        }

    if qtype:
        if qtype == 'slurm':
            hlp.pop('torque')
        elif qtype == 'torque':
            hlp.pop('slurm')
        else:
            raise ClusterError('qtype must be "torque", "slurm"')

    if mode == 'print' or mode == 'string':
        outstr = ''
        for hlp_info in hlp.values():
            tmpstr = ''
            for option, inf in hlp_info['help'].items():
                default   = inf['default'] if 'default' in inf else None
                typ       = inf['type']
                helpitems = _wrap(inf['help'])
                helpstr   = helpitems[0]
                if len(helpitems) > 1:
                    hstr     = '\n' + ' '*15
                    helpstr += hstr
                    helpstr += hstr.join(helpitems[1:])
                if isinstance(typ, (tuple, list, set)):
                    typ = [t.__name__ for t in typ]
                else:
                    typ = typ.__name__
                tmpstr += ('{o:<15}{h}\n{s:<15}Type: {t}; Default: {d}\n'
                           .format(o=option + ':', h=helpstr, s=' ',
                                   t=typ, d=default))
            outstr += '{}::\n{}\n'.format(hlp_info['summary'], tmpstr)
        outstr = outstr.rstrip() + '\n'

        if mode == 'print':
            _sys.stdout.write(outstr)
        else:
            return outstr

    elif mode == 'table':
        tables = _OD()
        for sect, ddct in hlp.items():
            summary = '{}: {}'.format(sect.title(), ddct['summary'])
            outtable = [['Option', 'Description', 'Type', 'Default']]
            dct = ddct['help']
            for opt, inf in dct.items():
                if isinstance(inf['type'], (tuple, list, set)):
                    typ = [t.__name__ for t in inf['type']]
                else:
                    typ = inf['type'].__name__
                outtable.append([
                    opt,
                    inf['help'],
                    typ,
                    str(inf['default'])
                ])
            tables[summary] = outtable

        tables['Synonyms'] = [
            ['Synonym', 'Option']
        ] + [
            list(i) for i in SYNONYMS.items()
        ]

        out_string = ''
        for section, table in tables.items():
            out_string += '\n' + section + '\n'
            out_string += '-'*len(section) + '\n\n'
            out_string += _tabulate(
                table, headers='firstrow', tablefmt=tablefmt
            ) + '\n\n'

        return out_string

    elif mode == 'merged_table':
        table = []
        headers  = ['Option', 'Description', 'Type', 'Default', 'Section']
        for sect, ddct in hlp.items():
            dct = ddct['help']
            for opt, inf in dct.items():
                if isinstance(inf['type'], (tuple, list, set)):
                    typ = [t.__name__ for t in inf['type']]
                else:
                    typ = inf['type'].__name__
                table.append([
                    opt,
                    inf['help'],
                    typ,
                    str(inf['default']),
                    sect
                ])
        out_string  = _tabulate(
            table, headers=headers, tablefmt=tablefmt
        ) + '\n\n'
        out_string += 'Synonyms\n'
        out_string += '-'*8 + '\n\n'
        out_string += _tabulate(
            [list(i) for i in SYNONYMS.items()],
            headers=['Synonym', 'Option'],
            tablefmt=tablefmt
        )

        return out_string

    elif mode == 'list':
        return '\n'.join(['\n'.join(i['help'].keys()) for i in hlp.values()])

    else:
        raise ClusterError('mode must be "print", "string", or "table"')
